import time
import warnings

from IPython import embed

import numpy

from scipy.special import erf
from scipy import interpolate

from matplotlib import pyplot, ticker, rc, colors, cm, colorbar, image

from astropy.io import fits

from mangadap import dapfits
from mangadap.datacube import DataCube
from mangadap.config import defaults
from mangadap.proc.util import growth_lim, inverse

from mangadap.config.analysisplan import AnalysisPlan
from mangadap.util.filter import BoxcarFilter
from mangadap.util.sampling import spectral_coordinate_step
from mangadap.util.mapping import map_extent, map_beam_patch
from mangadap.util.fileio import channel_dictionary
from mangadap.util.pkg import load_object

from mangadap.scripts import scriptbase


def init_ax(fig, pos, facecolor=None, grid=False):
    ax = fig.add_axes(pos, facecolor=facecolor)
    ax.minorticks_on()
    ax.tick_params(which='major', length=6) #, direction='in')
    ax.tick_params(which='minor', length=3) #, direction='in')
    if grid:
        ax.grid(True, which='major', color='0.8', zorder=1, linestyle='-', lw=0.5)
    return ax


def init_image_ax(fig, pos):
    ax = fig.add_axes(pos, facecolor='0.9')
    ax.minorticks_on()
    ax.grid(True, which='major', color='0.8', zorder=0, linestyle='-')
    ax.minorticks_on()
    ax.tick_params(which='major', length=6, direction='in')
    ax.tick_params(which='minor', length=3, direction='in')
    return ax


def masked_imshow(fig, ax, cax, data, extent=None, norm=None, vmin=None, vmax=None, cmap='viridis',
                  zorder=0, cbformat=None, subs=None):
    if numpy.sum(data.mask) != numpy.prod(data.shape):
        if norm is None:
            img = ax.imshow(data, origin='lower', interpolation='nearest', extent=extent,
                            vmin=vmin, vmax=vmax, cmap=cmap, zorder=zorder)
        else:
            img = ax.imshow(data, origin='lower', interpolation='nearest', extent=extent,
                            norm=norm, cmap=cmap, zorder=zorder)

        cb = fig.colorbar(img, cax=cax) if cbformat is None else \
                    fig.colorbar(img, cax=cax, format=cbformat)
        if subs is not None:
            cb.locator = ticker.LogLocator(base=10, subs=(1.,2.,4.,))
            cb.update_ticks()
    else:
        _norm = colors.Normalize(vmin=vmin, vmax=vmax) if norm is None else norm
        cb = colorbar.ColorbarBase(cax, cmap=cm.get_cmap(cmap), norm=_norm)


# TODO: Should make this a datacube member function
def gmr_data(cube, min_frac=0.8):
    """
    Read or construct a g-r for the provided datacube.
    """
    # First see if there are GIMG and RIMG extensions in the input datacube
    # (requires a MaNGA DRP-like datamodel)
    with fits.open(str(cube.file_path)) as hdu:
        ext_list = [h.name for h in hdu]
        if 'GIMG' in ext_list and 'RIMG' in ext_list:
            return -2.5*numpy.ma.log10(numpy.ma.MaskedArray(hdu['GIMG'].data,
                                                            mask=numpy.invert(hdu['GIMG'].data>0))
                                    / numpy.ma.MaskedArray(hdu['RIMG'].data,
                                                           mask=numpy.invert(hdu['RIMG'].data>0)))

    # Next, try to create it from scratch
    g_file = defaults.dap_data_root() / 'filter_response' / 'gunn_2001_g_response.db'
    if not g_file.exists():
        warnings.warn(f'Could not find file {g_file}; cannot produce rough g-r image.')
        return None

    r_file = defaults.dap_data_root() / 'filter_response' / 'gunn_2001_r_response.db'
    if not r_file.exists():
        warnings.warn(f'Could not find file {r_file}; cannot produce rough g-r image.')
        return None

    flux = cube.copy_to_masked_array()
    gpm = numpy.logical_not(numpy.ma.getmaskarray(flux)).astype(float)

    # Wavelengths in these files are in angstroms
    g_wave, g_eff = numpy.genfromtxt(str(g_file)).T
    indx = g_eff > 0.
    g_lo = numpy.amin(g_wave[indx])
    g_hi = numpy.amax(g_wave[indx])
    g_wgt = interpolate.interp1d(g_wave[indx], g_eff[indx], bounds_error=False,
                                 fill_value=0.)(cube.wave)[None,:]*gpm
    g_mask = numpy.sum((g_wgt[:, 1:]>0).astype(float) * numpy.diff(cube.wave)[None, :],
                       axis=1).reshape(cube.spatial_shape) < min_frac*(g_hi-g_lo)
    if numpy.all(g_mask):
        warnings.warn('Datacube has insufficient coverage of the g-band to construct a g-r image.')
        return None

    r_wave, r_eff = numpy.genfromtxt(str(r_file)).T
    indx = r_eff > 0.
    r_lo = numpy.amin(r_wave[indx])
    r_hi = numpy.amax(r_wave[indx])
    r_wgt = interpolate.interp1d(r_wave[indx], r_eff[indx], bounds_error=False,
                                 fill_value=0.)(cube.wave)[None,:]*gpm

    r_mask = numpy.sum((r_wgt[:, 1:]>0).astype(float) * numpy.diff(cube.wave)[None, :],
                       axis=1).reshape(cube.spatial_shape) < min_frac*(r_hi-r_lo)
    if numpy.all(r_mask):
        warnings.warn('Datacube has insufficient coverage of the r-band to construct a g-r image.')
        return None

    g = numpy.sum(g_wgt[:, 1:] * flux[:, 1:] * numpy.diff(cube.wave)[None,:],
                       axis=1).reshape(cube.spatial_shape)
    r = numpy.sum(r_wgt[:, 1:] * flux[:, 1:] * numpy.diff(cube.wave)[None,:],
                       axis=1).reshape(cube.spatial_shape)
    return -2.5*numpy.ma.log10(numpy.ma.divide(numpy.ma.MaskedArray(g, mask=g_mask),
                                               numpy.ma.MaskedArray(r, mask=r_mask)))


def get_stellar_continuum_data(dapmaps, dapcube):

    if not dapmaps.exists():
        raise FileNotFoundError(f'{dapmaps} does not exist.')
    if not dapcube.exists():
        raise FileNotFoundError(f'{dapcube} does not exist.')

    # Get the data
    cube_hdu = fits.open(str(dapcube))

    # Find the unique bins and get rid of -1 bins
    binid = cube_hdu['BINID'].data[1,:,:].ravel()
#    uniq, indx = map(lambda x: x[1:], numpy.unique(binid, return_index=True))
    uniq, indx = numpy.unique(binid, return_index=True)
    if uniq[0] == -1:
        uniq = uniq[1:]
        indx = indx[1:]

    # Get the unique fluxes, errors, and models
    npix = cube_hdu['FLUX'].data.shape[0]
    flux = numpy.ma.MaskedArray(cube_hdu['FLUX'].data.reshape(npix,-1),
                                mask=cube_hdu['MASK'].data.reshape(npix,-1) > 0)[:,indx]
    error = numpy.ma.power(numpy.ma.MaskedArray(cube_hdu['IVAR'].data.reshape(npix,-1),
                                                mask=cube_hdu['MASK'].data.reshape(npix,-1) > 0),
                           -0.5)[:,indx]
    model = numpy.ma.MaskedArray(cube_hdu['STELLAR'].data.reshape(npix,-1),
                                 mask=cube_hdu['STELLAR_MASK'].data.reshape(npix,-1) > 0)[:,indx]

    # Get the associated S/N and fit metrics
    maps_hdu = fits.open(str(dapmaps))
    snrg = maps_hdu['BIN_SNR'].data.ravel()[indx]
    rms = maps_hdu['STELLAR_FOM'].data[0,:,:].ravel()[indx]
    frms = maps_hdu['STELLAR_FOM'].data[1,:,:].ravel()[indx]
    rchi2 = maps_hdu['STELLAR_FOM'].data[2,:,:].ravel()[indx]

    return cube_hdu['WAVE'].data, flux, error, model, snrg, rms, frms, rchi2


def get_stellar_continuum_fom_maps(cube, dapmaps):

    if not dapmaps.exists():
        raise FileNotFoundError('{0} does not exist.'.format(dapmaps))

    # Get the associated S/N and fit metrics
    maps_hdu = fits.open(str(dapmaps))
    snrg = numpy.ma.MaskedArray(maps_hdu['BIN_SNR'].data,
                                mask=numpy.invert(maps_hdu['BIN_SNR'].data > 0))
    mask = maps_hdu['STELLAR_VEL_MASK'].data > 0
    rms = numpy.ma.MaskedArray(maps_hdu['STELLAR_FOM'].data[0,:,:], mask=mask)
    frms = numpy.ma.MaskedArray(maps_hdu['STELLAR_FOM'].data[1,:,:], mask=mask)
    rchi2 = numpy.ma.MaskedArray(maps_hdu['STELLAR_FOM'].data[2,:,:], mask=mask)
    chi_grw = numpy.ma.MaskedArray(maps_hdu['STELLAR_FOM'].data[6:,:,:])
    chi_grw[0,mask] = numpy.ma.masked
    chi_grw[1,mask] = numpy.ma.masked
    chi_grw[2,mask] = numpy.ma.masked

    extent = map_extent(maps_hdu, 'SPX_MFLUX')

    return extent, snrg, gmr_data(cube), rms, frms, rchi2, chi_grw


def get_emission_line_data(cube, daptype, dapmaps, dapcube):
    if not dapmaps.exists():
        raise FileNotFoundError('{0} does not exist.'.format(dapmaps))
    if not dapcube.exists():
        raise FileNotFoundError('{0} does not exist.'.format(dapcube))

    # Get the data
    cube_hdu = fits.open(str(dapcube))
    maps_hdu = fits.open(str(dapmaps))

    # Get the observed spectra and errors
    if 'HYB' in daptype:
        # Fits are to unbinned spectra from DRP datacube
        npix = cube.flux.shape[-1]
        flux = cube.copy_to_masked_array(attr='flux').T
        error = numpy.sqrt(inverse(cube.copy_to_masked_array(attr='ivar').T))
        snrg = maps_hdu['SPX_SNR'].data.ravel()

    else:
        # Fits are to binned spectra from DAP datacube
        npix = cube_hdu['FLUX'].data.shape[0]
        flux = numpy.ma.MaskedArray(cube_hdu['FLUX'].data.reshape(npix,-1),
                                    mask=cube_hdu['MASK'].data.reshape(npix,-1) > 0)
        error = numpy.ma.power(numpy.ma.MaskedArray(cube_hdu['IVAR'].data.reshape(npix,-1),
                                                mask=cube_hdu['MASK'].data.reshape(npix,-1) > 0),
                               -0.5)
        snrg = maps_hdu['BIN_SNR'].data.ravel()
   
    # Find the unique bins and get rid of -1 bins
    binid = cube_hdu['BINID'].data[3,:,:].ravel()
    uniq, indx = numpy.unique(binid, return_index=True)
    if uniq[0] == -1:
        uniq = uniq[1:]
        indx = indx[1:]

    flux = flux[:,indx]
    error = error[:,indx]
    model = numpy.ma.MaskedArray(cube_hdu['MODEL'].data.reshape(npix,-1),
                                 mask=cube_hdu['MODEL_MASK'].data.reshape(npix,-1) > 0)[:,indx]

    # Get the associated S/N and fit metrics
    snrg = snrg[indx]
    rms = maps_hdu['EMLINE_FOM'].data[0,:,:].ravel()[indx]
    frms = maps_hdu['EMLINE_FOM'].data[1,:,:].ravel()[indx]
    rchi2 = maps_hdu['EMLINE_FOM'].data[2,:,:].ravel()[indx]

    return cube_hdu['WAVE'].data, flux, error, model, snrg, rms, frms, rchi2


def get_emission_line_fom_maps(dapmaps):

    maps_hdu = fits.open(dapmaps)
    eml = channel_dictionary(maps_hdu, 'EMLINE_GFLUX')
    anr = numpy.ma.MaskedArray(maps_hdu['EMLINE_GANR'].data[eml['Ha-6564'],...],
                               mask=maps_hdu['EMLINE_GFLUX_MASK'].data[eml['Ha-6564'],...] > 0)
    ha = numpy.ma.MaskedArray(maps_hdu['EMLINE_GFLUX'].data[eml['Ha-6564'],...],
                              mask=maps_hdu['EMLINE_GFLUX_MASK'].data[eml['Ha-6564'],...] > 0)
    hb = numpy.ma.MaskedArray(maps_hdu['EMLINE_GFLUX'].data[eml['Hb-4862'],...],
                              mask=maps_hdu['EMLINE_GFLUX_MASK'].data[eml['Hb-4862'],...] > 0)
    aob = numpy.ma.divide(ha, hb)

    # Get the associated S/N and fit metrics
    rms = numpy.ma.MaskedArray(maps_hdu['EMLINE_FOM'].data[0,:,:])
    rms[numpy.invert(rms > 0)] = numpy.ma.masked
    frms = numpy.ma.MaskedArray(maps_hdu['EMLINE_FOM'].data[1,:,:])
    frms[numpy.invert(frms > 0)] = numpy.ma.masked
    rchi2 = numpy.ma.MaskedArray(maps_hdu['EMLINE_FOM'].data[2,:,:])
    rchi2[numpy.invert(rchi2 > 0)] = numpy.ma.masked
    chi_grw = numpy.ma.MaskedArray(maps_hdu['EMLINE_FOM'].data[6:,:,:])
    chi_grw[numpy.invert(chi_grw > 0)] = numpy.ma.masked

    extent = map_extent(maps_hdu, 'SPX_MFLUX')

    return extent, anr, aob, rms, frms, rchi2, chi_grw


def fom_lambda(name, wave, flux, error, model, fit='sc', wave_limits=None, ofile=None):

    if wave_limits is not None:
        indx = (wave < wave_limits[0]) | (wave > wave_limits[1])
        flux[indx,:] = numpy.ma.masked
        error[indx,:] = numpy.ma.masked
        model[indx,:] = numpy.ma.masked
    else:
        indx = numpy.arange(len(wave))[numpy.any(numpy.invert(model.mask), axis=1)]
        wave_limits = [wave[indx[0]], wave[indx[-1]]]

    nspec = flux.shape[1]
    resid = numpy.ma.absolute(flux-model)
    fresid = numpy.ma.absolute(numpy.ma.divide(resid, model))
    chi = numpy.ma.absolute(numpy.ma.divide(resid,error))
    snr = numpy.ma.divide(flux, error)

    mean_flux = numpy.ma.mean(flux, axis=1)
    mean_snr = numpy.ma.mean(snr, axis=1)
    mean_resid = numpy.ma.mean(resid, axis=1)
    mean_fresid = numpy.ma.mean(fresid, axis=1)
    mean_chi = numpy.ma.mean(chi, axis=1)

    bc = BoxcarFilter(100)
    resid_sm = bc.smooth(resid)
    fresid_sm = bc.smooth(fresid)
    chi_sm = bc.smooth(chi)

    spec_lim = [-10, nspec+10]

    flux_lim = numpy.exp(growth_lim(numpy.ma.log(mean_flux).compressed(), 0.99, fac=2.0))
    snr_lim = numpy.exp(growth_lim(numpy.ma.log(mean_snr).compressed(), 0.99, fac=2.0))
    resid_lim = numpy.exp(growth_lim(numpy.ma.log(mean_resid).compressed(), 0.99, fac=2.0))
    fresid_lim = numpy.exp(growth_lim(numpy.ma.log(mean_fresid).compressed(), 0.99, fac=2.0))
    chi_lim = numpy.exp(growth_lim(numpy.ma.log(mean_chi).compressed(), 0.99, fac=2.0, midpoint=0))

    l0 = numpy.log10(wave[0])
    dl = spectral_coordinate_step(wave, log=True)
    wave_bins = numpy.power(10, l0-dl/2 + dl*numpy.arange(len(wave)+1))
    spec_bins = 0.5+numpy.arange(nspec+1)

    font = { 'size' : 10 }
    rc('font', **font)

    w,h = pyplot.figaspect(1)
    fig = pyplot.figure(figsize=(1.5*w,1.5*h))

    ax = init_ax(fig, [0.2, 0.87, 0.60, 0.12], facecolor='0.95', grid=True)
    ax.set_xlim(wave_limits)
    ax.set_ylim(flux_lim)
    ax.set_yscale('log')
    ax.xaxis.set_major_formatter(ticker.NullFormatter())
    ax.plot(wave, numpy.ma.mean(flux, axis=1), color='k', lw=0.5)
    ax.text(-0.15, 0.5, r'Flux', horizontalalignment='center',
            verticalalignment='center', rotation='vertical', transform=ax.transAxes)
    axt = ax.twinx()
    axt.set_xlim(wave_limits)
    axt.set_ylim(snr_lim)
    axt.set_yscale('log')
    axt.tick_params(axis='y', colors='0.5')
    axt.xaxis.set_major_formatter(ticker.NullFormatter())
    axt.plot(wave, numpy.ma.mean(snr, axis=1), color='0.5', lw=0.5)
    axt.text(1.1, 0.5, r'S/N', horizontalalignment='center',
            verticalalignment='center', rotation='vertical', transform=axt.transAxes,
            color='0.5')

    ax = init_ax(fig, [0.2, 0.75, 0.60, 0.12], facecolor='0.95', grid=True)
    ax.set_xlim(wave_limits)
    ax.set_ylim(resid_lim)
    ax.set_yscale('log')
    ax.xaxis.set_major_formatter(ticker.NullFormatter())
#    ax.plot(wave, numpy.ma.mean(resid_sm, axis=0), color='C3')
    ax.plot(wave, numpy.ma.mean(resid, axis=1), color='k', lw=0.5)
    ax.text(-0.15, 0.5, r'$|\Delta|$', horizontalalignment='center',
            verticalalignment='center', rotation='vertical', transform=ax.transAxes)

    ax = init_ax(fig, [0.2, 0.63, 0.60, 0.12], facecolor='0.95', grid=True)
    ax.set_xlim(wave_limits)
    ax.set_ylim(fresid_lim)
    ax.set_yscale('log')
    ax.xaxis.set_major_formatter(ticker.NullFormatter())
    ax.plot(wave, numpy.ma.mean(fresid, axis=1), color='k', lw=0.5)
    ax.text(-0.15, 0.5, r'$|\Delta|/m$', horizontalalignment='center',
            verticalalignment='center', rotation='vertical', transform=ax.transAxes)

    ax = init_ax(fig, [0.2, 0.51, 0.60, 0.12], facecolor='0.95', grid=True)
    ax.set_xlim(wave_limits)
    ax.set_ylim(chi_lim)
    ax.set_yscale('log')
    ax.plot(wave, numpy.ma.mean(chi, axis=1), color='k', lw=0.5)
    ax.text(0.5, -0.41, r'Wavelength ($\AA$)', horizontalalignment='center',
            verticalalignment='center', transform=ax.transAxes)
    ax.text(-0.15, 0.5, r'$|\Delta|/\epsilon$', horizontalalignment='center',
            verticalalignment='center', rotation='vertical', transform=ax.transAxes)

    tracer = 'STELLAR' if fit == 'sc' else 'EMLINE'
    ax.text(1.23, -0.3, f'{name} {tracer}', horizontalalignment='center',
            verticalalignment='bottom', rotation='vertical', transform=ax.transAxes,
            fontsize=22)

    ax = init_ax(fig, [0.07, 0.23, 0.30, 0.16], facecolor='0.95')
    ax.set_xlim(wave_limits)
    ax.set_ylim(spec_lim)
    ax.xaxis.set_major_formatter(ticker.NullFormatter())
    im = ax.pcolormesh(wave_bins, spec_bins, resid.T,
                       norm=colors.LogNorm(vmin=resid_lim[0], vmax=resid_lim[1]),
                       cmap='viridis', zorder=4, lw=0, rasterized=True)
    ax = init_ax(fig, [0.07, 0.07, 0.30, 0.16], facecolor='0.95')
    ax.set_xlim(wave_limits)
    ax.set_ylim(spec_lim)
    ax.pcolormesh(wave_bins, spec_bins, resid_sm.T,
                  norm=colors.LogNorm(vmin=resid_lim[0], vmax=resid_lim[1]),
                  cmap='viridis', zorder=4, lw=0, rasterized=True)
    cax = fig.add_axes([0.07, 0.395, 0.20, 0.01])
    cb = pyplot.colorbar(im, cax=cax, orientation='horizontal') #, format=FormatStrFormatter('%d'))
    cb.ax.tick_params(axis='x', which='both', bottom=False, top=True, labelbottom=False,
                      labeltop=True)
    cax.text(1.05, 0.7, r'$|\Delta|$', horizontalalignment='left', verticalalignment='center',
            transform=cax.transAxes, fontsize=10)
    
    ax = init_ax(fig, [0.37, 0.23, 0.30, 0.16], facecolor='0.95')
    ax.set_xlim(wave_limits)
    ax.set_ylim(spec_lim)
    ax.xaxis.set_major_formatter(ticker.NullFormatter())
    ax.yaxis.set_major_formatter(ticker.NullFormatter())
    im = ax.pcolormesh(wave_bins, spec_bins, fresid.T,
                       norm=colors.LogNorm(vmin=fresid_lim[0], vmax=fresid_lim[1]),
                       cmap='viridis', zorder=4, lw=0, rasterized=True)
    ax = init_ax(fig, [0.37, 0.07, 0.30, 0.16], facecolor='0.95')
    ax.set_xlim(wave_limits)
    ax.set_ylim(spec_lim)
    ax.yaxis.set_major_formatter(ticker.NullFormatter())
    ax.pcolormesh(wave_bins, spec_bins, fresid_sm.T,
                  norm=colors.LogNorm(vmin=fresid_lim[0], vmax=fresid_lim[1]),
                  cmap='viridis', zorder=4, lw=0, rasterized=True)
    ax.text(0.5, -0.30, r'Wavelength ($\AA$)', horizontalalignment='center',
            verticalalignment='center', transform=ax.transAxes)
    cax = fig.add_axes([0.37, 0.395, 0.20, 0.01])
    cb = pyplot.colorbar(im, cax=cax, orientation='horizontal') #, format=FormatStrFormatter('%d'))
    cb.ax.tick_params(axis='x', which='both', bottom=False, top=True, labelbottom=False,
                      labeltop=True)
    cax.text(1.05, 0.7, r'$|\Delta|/m$', horizontalalignment='left', verticalalignment='center',
            transform=cax.transAxes, fontsize=10)

    ax = init_ax(fig, [0.67, 0.23, 0.30, 0.16], facecolor='0.95')
    ax.set_xlim(wave_limits)
    ax.set_ylim(spec_lim)
    ax.xaxis.set_major_formatter(ticker.NullFormatter())
    ax.yaxis.set_major_formatter(ticker.NullFormatter())
    im = ax.pcolormesh(wave_bins, spec_bins, chi.T,
                       norm=colors.LogNorm(vmin=chi_lim[0], vmax=chi_lim[1]),
                       cmap='RdBu_r', zorder=4, lw=0, rasterized=True)
    ax = init_ax(fig, [0.67, 0.07, 0.30, 0.16], facecolor='0.95')
    ax.set_xlim(wave_limits)
    ax.set_ylim(spec_lim)
    ax.yaxis.set_major_formatter(ticker.NullFormatter())
    ax.pcolormesh(wave_bins, spec_bins, chi_sm.T,
                  norm=colors.LogNorm(vmin=chi_lim[0], vmax=chi_lim[1]),
                  cmap='RdBu_r', zorder=4, lw=0, rasterized=True)
    cax = fig.add_axes([0.67, 0.395, 0.20, 0.01])
    cb = pyplot.colorbar(im, cax=cax, orientation='horizontal') #, format=FormatStrFormatter('%d'))
    cb.ax.tick_params(axis='x', which='both', bottom=False, top=True, labelbottom=False,
                      labeltop=True)
    cax.text(1.05, 0.7, r'$|\Delta|/\epsilon$', horizontalalignment='left',
             verticalalignment='center', transform=cax.transAxes, fontsize=10)

    if ofile is None:
        pyplot.show()
    else:
        print('Writing: {0}'.format(ofile))
        fig.canvas.print_figure(ofile, bbox_inches='tight', dpi=100)
    fig.clear()
    pyplot.close(fig)


def fom_growth(name, flux, error, model, snrg, rms, frms, rchi2, fit='sc', ofile=None):

    nspec = flux.shape[1]
    resid = numpy.ma.absolute(flux-model)
    fresid = numpy.ma.divide(resid, model)
    chi = numpy.ma.divide(resid, error)

    snr_lim = numpy.exp(growth_lim(numpy.ma.log(snrg).compressed(), 0.95, fac=1.3))

    chi_lim = numpy.exp(growth_lim(numpy.ma.log(chi).compressed(), 0.8, fac=1.1))
    chi_lim[1] = numpy.ma.amax(chi)*2
    resid_lim = numpy.exp(growth_lim(numpy.ma.log(resid).compressed(), 0.9, fac=1.1))
    resid_lim[1] = numpy.ma.amax(resid)*2
    fresid_lim = numpy.exp(growth_lim(numpy.ma.log(fresid).compressed(), 0.9, fac=1.1))
    fresid_lim[1] = numpy.ma.amax(fresid)*2

    w,h = pyplot.figaspect(1)
    fig = pyplot.figure(figsize=(1.5*w,1.5*h))

    rax = init_ax(fig, [0.07, 0.50, 0.30, 0.35], facecolor='0.9', grid=True)
    rax.set_xlim(resid_lim)
    rax.set_ylim([1e-4,2])
    rax.set_yscale('log')
    rax.set_xscale('log')
    rax.text(0.5, -0.17, r'$|\Delta|$', ha='center', va='center', transform=rax.transAxes)
    rax.text(0.1, 0.5, r'1-Growth', horizontalalignment='center',
             verticalalignment='center', transform=rax.transAxes, rotation='vertical')

    fax = init_ax(fig, [0.37, 0.50, 0.30, 0.35], facecolor='0.9', grid=True)
    fax.set_xlim(fresid_lim)
    fax.set_ylim([1e-4,2])
    fax.set_yscale('log')
    fax.set_xscale('log')
    fax.yaxis.set_major_formatter(ticker.NullFormatter())
    tracer = 'STELLAR' if fit == 'sc' else 'EMLINE'
    fax.text(0.5, 1.3, f'{name} {tracer}', horizontalalignment='center',
             verticalalignment='center', transform=fax.transAxes, fontsize=22)
    fax.text(0.5, -0.17, r'$|\Delta|/m$', ha='center', va='center', transform=fax.transAxes)

    cax = init_ax(fig, [0.67, 0.50, 0.30, 0.35], facecolor='0.9', grid=True)
    cax.set_xlim(chi_lim)
    cax.set_ylim([1e-4,2])
    cax.set_yscale('log')
    cax.set_xscale('log')
    cax.yaxis.set_major_formatter(ticker.NullFormatter())
    cax.text(0.5, -0.17, r'$|\Delta|/\epsilon$', ha='center', va='center',
             transform=cax.transAxes)


    cmap = cm.get_cmap('viridis')
    cnorm = colors.LogNorm(vmin=snr_lim[0], vmax=snr_lim[1])
    cbax = fig.add_axes([0.69, 0.86, 0.2, 0.01])

    cb = colorbar.ColorbarBase(cbax, norm=cnorm, cmap=cmap, orientation='horizontal')
    cb.ax.tick_params(axis='x', which='both', bottom=False, top=True, labelbottom=False,
                      labeltop=True)
    cbax.text(1.05, 0.5, r'S/N$_g$', ha='left', va='center', transform=cbax.transAxes)

    snr_srt = numpy.argsort(snrg)

    for i in range(nspec):
        j = snr_srt[i]
        if numpy.all(chi.mask[:,j]):
            continue

        t = resid[:,j].compressed()
        s = numpy.argsort(t)
        rax.step(t[s], 1-numpy.arange(len(t)).astype(float)/len(t), where='pre',
                 color=cmap(cnorm(snrg[j])), lw=0.7)

        t = fresid[:,j].compressed()
        s = numpy.argsort(t)
        fax.step(t[s], 1-numpy.arange(len(t)).astype(float)/len(t), where='pre',
                 color=cmap(cnorm(snrg[j])), lw=0.7)

        t = chi[:,j].compressed()
        s = numpy.argsort(t)
        cax.step(t[s], 1-numpy.arange(len(t)).astype(float)/len(t), where='pre',
                 color=cmap(cnorm(snrg[j])), lw=0.7)

    x = numpy.linspace(0,5,100)
    g = (erf(x/numpy.sqrt(2)) - erf(-x/numpy.sqrt(2)))/2.
    interp = interpolate.interp1d(g, x)
    cax.plot(x,1-g, color='k')

    chi2_lim = numpy.exp(growth_lim(numpy.ma.log(rchi2).compressed(), 1.0, fac=2.0))
    rms_lim = numpy.exp(growth_lim(numpy.ma.log(rms).compressed(), 1.0, fac=2.0))
    frms_lim = numpy.exp(growth_lim(numpy.ma.log(frms).compressed(), 1.0, fac=2.0))
    snr_lim = numpy.exp(growth_lim(numpy.ma.log(snrg).compressed(), 1.0, fac=2.0))

    rax = init_ax(fig, [0.07, 0.15, 0.30, 0.25], facecolor='0.9', grid=True)
    rax.set_xlim(rms_lim)
    rax.set_ylim(snr_lim)
    rax.set_yscale('log')
    rax.set_xscale('log')
    rax.scatter(rms, snrg,
                marker='.', s=20, color='k', lw=0, alpha=0.5)
    rax.text(0.5, -0.25, r'$RMS$', ha='center', va='center', transform=rax.transAxes)
    rax.text(0.1, 0.5, r'S/N$_g$', horizontalalignment='center',
             verticalalignment='center', transform=rax.transAxes, rotation='vertical')

    fax = init_ax(fig, [0.37, 0.15, 0.30, 0.25], facecolor='0.9', grid=True)
    fax.set_xlim(frms_lim)
    fax.set_ylim(snr_lim)
    fax.set_yscale('log')
    fax.set_xscale('log')
    fax.yaxis.set_major_formatter(ticker.NullFormatter())
    fax.scatter(frms, snrg,
                marker='.', s=20, color='k', lw=0, alpha=0.5)
    fax.text(0.5, -0.25, r'$fRMS$', ha='center', va='center', transform=fax.transAxes)

    cax = init_ax(fig, [0.67, 0.15, 0.30, 0.25], facecolor='0.9', grid=True)
    cax.set_xlim(chi2_lim)
    cax.set_ylim(snr_lim)
    cax.set_yscale('log')
    cax.set_xscale('log')
    cax.yaxis.set_major_formatter(ticker.NullFormatter())
    cax.scatter(rchi2, snrg,
                marker='.', s=20, color='k', lw=0, alpha=0.5)
    cax.text(0.5, -0.25, r'$\chi^2_\nu$', ha='center', va='center', transform=cax.transAxes)
    
    if ofile is None:
        pyplot.show()
    else:
        print('Writing: {0}'.format(ofile))
        fig.canvas.print_figure(ofile, bbox_inches='tight', dpi=100)
    fig.clear()
    pyplot.close(fig)


def fom_maps(name, image_file, snr, color, rms, frms, rchi2, chi_grw, fwhm=None, extent=None,
             fit='sc', ofile=None):

    if color is not None:
        color_lim = numpy.exp(growth_lim(numpy.ma.log(color), 0.95, fac=1.1)) \
                        if fit != 'sc' else growth_lim(color, 0.95, fac=1.10)

    snr_lim = numpy.power(10., growth_lim(numpy.ma.log10(snr), 0.90, fac=1.05))
    rms_lim = numpy.power(10., growth_lim(numpy.ma.log10(rms), 0.90, fac=1.05))
    frm_lim = numpy.power(10., growth_lim(numpy.ma.log10(frms), 0.90, fac=1.05))

    chi_lim = [1/4, 4]
    c68_lim = [1/4, 4]
    c99_lim = [2.6/4, 2.6*4]
    crt_lim = [0.8, 2]

    # Get the limits to apply
    x_order = 1 if extent[0] < extent[1] else -1
    Dx = x_order*(extent[1] - extent[0])
    y_order = 1 if extent[2] < extent[3] else -1
    Dy = y_order*(extent[3] - extent[2])
    # Force the image panels to be square; assumes extent has the same units in
    # both dimensions.
    D = max(Dx, Dy)

    x_lim = (extent[0]+extent[1])/2 + D * x_order * numpy.array([-1,1])/2
    y_lim = (extent[2]+extent[3])/2 + D * y_order * numpy.array([-1,1])/2

    left = 0.035
    bott = 0.08
    imwd = 0.24
    hbuf = 0.08
    cbuf = 0.005
    cbwd = 0.01
    vbuf = 0.03

    font = { 'size' : 6 }
    rc('font', **font)

    w,h = pyplot.figaspect(1)
    fig = pyplot.figure(figsize=(1.5*w,1.5*h))

    i,j=0,2
    ax = init_image_ax(fig, [left+i*(imwd+hbuf), bott+j*(imwd+vbuf), imwd, imwd ])
    if image_file is not None and image_file.exists():
        img = image.imread(str(image_file))
        ax.imshow(img)
        ax.text(0.95, 0.95, r'SDSS', color='white', ha='right', va='center', transform=ax.transAxes)
    else:
        ax.text(0.5, 0.5, 'No Image', horizontalalignment='center', verticalalignment='center',
                transform=ax.transAxes, fontsize=20)
    ax.axes.get_xaxis().set_visible(False)
    ax.axes.get_yaxis().set_visible(False)

    i,j=1,2
    ax = init_image_ax(fig, [left+i*(imwd+hbuf), bott+j*(imwd+vbuf), imwd, imwd ])
    cax = fig.add_axes([left+i*(imwd+hbuf)+imwd+cbuf, bott+j*(imwd+vbuf), cbwd, imwd ])
    masked_imshow(fig, ax, cax, snr, extent=extent,
                  norm=colors.LogNorm(vmin=snr_lim[0], vmax=snr_lim[1]), cmap='viridis',
                  zorder=3)
    ax.set_xlim(x_lim)
    ax.set_ylim(y_lim)
    if fwhm is not None:
        ax.add_patch(map_beam_patch(extent, ax, fwhm=fwhm, facecolor='0.7', edgecolor='k', zorder=4))
    ax.text(0.95, 0.95, r'S/N$_g$' if fit == 'sc' else r'H$\alpha$ A/N', ha='right', va='center',
            transform=ax.transAxes, zorder=4)

    tracer = 'STELLAR' if fit == 'sc' else 'EMLINE'
    ax.text(0.5, 1.1, f'{name} {tracer}', horizontalalignment='center',
            verticalalignment='bottom', transform=ax.transAxes, fontsize=26)

    i,j=2,2
    ax = init_image_ax(fig, [left+i*(imwd+hbuf), bott+j*(imwd+vbuf), imwd, imwd ])
    if color is not None:
        cax = fig.add_axes([left+i*(imwd+hbuf)+imwd+cbuf, bott+j*(imwd+vbuf), cbwd, imwd ])
        kwargs = {}
        if fit == 'sc':
            # Plotting the g-r color
            kwargs['vmin'] = color_lim[0]
            kwargs['vmax'] = color_lim[1]
        else:
            # Plotting H-alpha/H-beta flux ratio
            kwargs['norm'] = colors.LogNorm(vmin=color_lim[0], vmax=color_lim[1])
        masked_imshow(fig, ax, cax, color, extent=extent, cmap='RdBu_r', zorder=3, **kwargs)
    else:
        ax.text(0.5, 0.5, 'No g-r data', ha='center', va='center', transform=ax.transAxes)
    ax.set_xlim(x_lim)
    ax.set_ylim(y_lim)
    if fwhm is not None:
        ax.add_patch(map_beam_patch(extent, ax, fwhm=fwhm, facecolor='0.7', edgecolor='k', zorder=4))
    ax.text(0.95, 0.95, r'$g-r$' if fit == 'sc' else r'H$\alpha$/H$\beta$', ha='right',
            va='center', transform=ax.transAxes, zorder=4)

    i,j=0,1
    ax = init_image_ax(fig, [left+i*(imwd+hbuf), bott+j*(imwd+vbuf), imwd, imwd ])
    cax = fig.add_axes([left+i*(imwd+hbuf)+imwd+cbuf, bott+j*(imwd+vbuf), cbwd, imwd ])
    masked_imshow(fig, ax, cax, rms, extent=extent,
                  norm=colors.LogNorm(vmin=rms_lim[0], vmax=rms_lim[1]), cmap='viridis',
                  zorder=3)
    ax.set_xlim(x_lim)
    ax.set_ylim(y_lim)
    if fwhm is not None:
        ax.add_patch(map_beam_patch(extent, ax, fwhm=fwhm, facecolor='0.7', edgecolor='k', zorder=4))
    ax.text(0.95, 0.95, r'RMS', ha='right', va='center', transform=ax.transAxes, zorder=4)

    i,j=1,1
    ax = init_image_ax(fig, [left+i*(imwd+hbuf), bott+j*(imwd+vbuf), imwd, imwd ])
    cax = fig.add_axes([left+i*(imwd+hbuf)+imwd+cbuf, bott+j*(imwd+vbuf), cbwd, imwd ])
    masked_imshow(fig, ax, cax, frms, extent=extent,
                  norm=colors.LogNorm(vmin=frm_lim[0], vmax=frm_lim[1]), cmap='viridis',
                  zorder=3)
    ax.set_xlim(x_lim)
    ax.set_ylim(y_lim)
    if fwhm is not None:
        ax.add_patch(map_beam_patch(extent, ax, fwhm=fwhm, facecolor='0.7', edgecolor='k', zorder=4))
    ax.text(0.95, 0.95, r'fRMS', ha='right', va='center', transform=ax.transAxes, zorder=4)

    i,j=2,1
    ax = init_image_ax(fig, [left+i*(imwd+hbuf), bott+j*(imwd+vbuf), imwd, imwd ])
    cax = fig.add_axes([left+i*(imwd+hbuf)+imwd+cbuf, bott+j*(imwd+vbuf), cbwd, imwd ])
    masked_imshow(fig, ax, cax, rchi2, extent=extent,
                  norm=colors.LogNorm(vmin=chi_lim[0], vmax=chi_lim[1]), cmap='RdBu_r',
                  zorder=3)
    ax.set_xlim(x_lim)
    ax.set_ylim(y_lim)
    if fwhm is not None:
        ax.add_patch(map_beam_patch(extent, ax, fwhm=fwhm, facecolor='0.7', edgecolor='k', zorder=4))
    ax.text(0.95, 0.95, r'$\chi^2$', ha='right', va='center', transform=ax.transAxes, zorder=4)

    i,j=0,0
    ax = init_image_ax(fig, [left+i*(imwd+hbuf), bott+j*(imwd+vbuf), imwd, imwd ])
    cax = fig.add_axes([left+i*(imwd+hbuf)+imwd+cbuf, bott+j*(imwd+vbuf), cbwd, imwd ])
    masked_imshow(fig, ax, cax, chi_grw[0], extent=extent,
                  norm=colors.LogNorm(vmin=c68_lim[0], vmax=c68_lim[1]), cmap='viridis',
                  zorder=3)
    ax.set_xlim(x_lim)
    ax.set_ylim(y_lim)
    if fwhm is not None:
        ax.add_patch(map_beam_patch(extent, ax, fwhm=fwhm, facecolor='0.7', edgecolor='k', zorder=4))
    ax.text(0.95, 0.95, r'68% Growth $|\Delta|/\epsilon$', ha='right', va='center',
            transform=ax.transAxes, zorder=4)

    i,j=1,0
    ax = init_image_ax(fig, [left+i*(imwd+hbuf), bott+j*(imwd+vbuf), imwd, imwd ])
    cax = fig.add_axes([left+i*(imwd+hbuf)+imwd+cbuf, bott+j*(imwd+vbuf), cbwd, imwd ])
    masked_imshow(fig, ax, cax, chi_grw[1], extent=extent,
                  norm=colors.LogNorm(vmin=c99_lim[0], vmax=c99_lim[1]), cmap='viridis',
                  zorder=3)
    ax.set_xlim(x_lim)
    ax.set_ylim(y_lim)
    if fwhm is not None:
        ax.add_patch(map_beam_patch(extent, ax, fwhm=fwhm, facecolor='0.7', edgecolor='k', zorder=4))
    ax.text(0.95, 0.95, r'99% Growth $|\Delta|/\epsilon$', ha='right', va='center',
            transform=ax.transAxes, zorder=4)

    x = numpy.linspace(0,5,100)
    g = (erf(x/numpy.sqrt(2)) - erf(-x/numpy.sqrt(2)))/2.
    interp = interpolate.interp1d(g, x)
    m = interp([0.68, 0.99])
    i,j=2,0
    ax = init_image_ax(fig, [left+i*(imwd+hbuf), bott+j*(imwd+vbuf), imwd, imwd ])
    cax = fig.add_axes([left+i*(imwd+hbuf)+imwd+cbuf, bott+j*(imwd+vbuf), cbwd, imwd ])
    masked_imshow(fig, ax, cax, numpy.ma.divide(chi_grw[1],chi_grw[0])*m[0]/m[1],
                  extent=extent, vmin=crt_lim[0], vmax=crt_lim[1], cmap='RdBu_r',
                  zorder=3)
    ax.set_xlim(x_lim)
    ax.set_ylim(y_lim)
    if fwhm is not None:
        ax.add_patch(map_beam_patch(extent, ax, fwhm=fwhm, facecolor='0.7', edgecolor='k', zorder=4))
    ax.text(0.95, 0.95, r'99%/68% wrt Gaussian', ha='right', va='center', transform=ax.transAxes,
            zorder=4)

    if ofile is None:
        pyplot.show()
    else:
        print('Writing: {0}'.format(ofile))
        fig.canvas.print_figure(ofile, bbox_inches='tight', dpi=100)
    fig.clear()
    pyplot.close(fig)


#def fit_residuals(cube, plan, method_dirdrpver, redux_path, dapver, analysis_path, daptype, plt, ifu):
def fit_residuals(cube, plan, method_dir, ref_dir, qa_dir, fwhm=None):

    if not method_dir.exists():
        raise NotADirectoryError(f'{method_dir} does not exist; run the DAP first!')
    if not ref_dir.exists():
        raise NotADirectoryError(f'{ref_dir} does not exist; run the DAP first!')
    if not qa_dir.exists():
        qa_dir.mkdir(parents=True)

    _, directory, file = dapfits.default_paths(cube, method=plan['key'], output_path=method_dir)
    maps_file = directory / file
    if not maps_file.exists():
        raise FileNotFoundError(f'No file: {maps_file}')

    image_file = cube.image_path if hasattr(cube, 'image_path') else None
    _, directory, file = dapfits.default_paths(cube, file_type='LOGCUBE', method=plan['key'],
                                               output_path=method_dir)
    cube_file = directory / file
    if not cube_file.exists():
        raise FileNotFoundError(f'No file: {cube_file}')

    # Stellar continuum fit qa plot
    ofile_root = f'{cube.output_root}-LOGCUBE-{plan["key"]}-sc-fitqa'
    wave, flux, error, model, snrg, rms, frms, rchi2 \
            = get_stellar_continuum_data(maps_file, cube_file)
    ofile = qa_dir / f'{ofile_root}-lambda.png'
    fom_lambda(cube.name, wave, flux, error, model, fit='sc', ofile=ofile)
    ofile = qa_dir / f'{ofile_root}-growth.png'
    fom_growth(cube.name, flux, error, model, snrg, rms, frms, rchi2, fit='sc', ofile=ofile)
    extent, snrg, gmr, rms, frms, rchi2, chi_grw \
            = get_stellar_continuum_fom_maps(cube, maps_file)
    ofile = qa_dir / f'{ofile_root}-maps.png'
    fom_maps(cube.name, image_file, snrg, gmr, rms, frms, rchi2, chi_grw, fwhm=fwhm, extent=extent,
             fit='sc', ofile=ofile)

    # Emission-line fit qa plot
    ofile_root = f'{cube.output_root}-LOGCUBE-{plan["key"]}-el-fitqa'
    wave, flux, error, model, snrg, rms, frms, rchi2 \
            = get_emission_line_data(cube, plan["key"], maps_file, cube_file)
    ofile = qa_dir / f'{ofile_root}-lambda.png'
    fom_lambda(cube.name, wave, flux, error, model, fit='el', ofile=ofile)
    ofile = qa_dir / f'{ofile_root}-growth.png'
    fom_growth(cube.name, flux, error, model, snrg, rms, frms, rchi2, fit='el', ofile=ofile)
    extent, anr, aob, rms, frms, rchi2, chi_grw \
            = get_emission_line_fom_maps(maps_file)
    ofile = qa_dir / f'{ofile_root}-maps.png'
    fom_maps(cube.name, image_file, anr, aob, rms, frms, rchi2, chi_grw, fwhm=fwhm, extent=extent,
             fit='el', ofile=ofile)

class FitResiduals(scriptbase.ScriptBase):

    @classmethod
    def get_parser(cls, width=None):

        parser = super().get_parser(description='Construct QA plots showing the fit residuals',
                                    width=width)

        group = parser.add_mutually_exclusive_group(required=True)
        group.add_argument('-c', '--config', type=str, default=None,
                           help='Configuration file used to instantiate the relevant DataCube '
                                'derived class.')
        group.add_argument('-f', '--cubefile', type=str, default=None,
                           help='Name of the file with the datacube data.  Must be possible to '
                                'instantiate the relevant DataCube derived class directly from '
                                'the file only.')
        parser.add_argument('--cube_module', nargs='*',
                            default='mangadap.datacube.MaNGADataCube',
                            help='The name of the module that contains the DataCube derived '
                                 'class used to read the data.')

        parser.add_argument('-p', '--plan', type=str,
                            help='SDSS parameter file with analysis plan.  If not provided, a '
                                 'default plan is used.')
        parser.add_argument('--plan_module', nargs='*',
                            default='mangadap.config.manga.MaNGAAnalysisPlan',
                            help='The name of the module used to define the analysis plan and '
                                 'the output paths.')

        parser.add_argument('-o', '--output_path', type=str, default=None,
                            help='Top-level directory for the DAP output files; default path is '
                                 'set by the provided analysis plan object (see plan_module).')

        parser.add_argument('-b', '--beam', type=float, default=None, help='Beam FWHM for plot.')

#        parser.add_argument('plate', type=int, help='plate ID to process')
#        parser.add_argument('ifudesign', type=int, help='IFU design to process')
#
#        parser.add_argument('--drpver', type=str, help='DRP version', default=None)
#        parser.add_argument('--dapver', type=str, help='DAP version', default=None)
#        parser.add_argument("--redux_path", type=str, help="main DRP output path", default=None)
#        parser.add_argument("--analysis_path", type=str, help="main DAP output path", default=None)
#
#        parser.add_argument("--plan_file", type=str, help="parameter file with the MaNGA DAP "
#                            "execution plan to use instead of the default" , default=None)
#
#        parser.add_argument('--daptype', type=str, help='DAP processing type', default=None)
        parser.add_argument('--normal_backend', dest='bgagg', action='store_false', default=True)

        return parser

    @staticmethod
    def main(args):
        t = time.perf_counter()

        if args.bgagg:
            pyplot.switch_backend('agg')

        # Instantiate the DataCube
        #   - Import the module used to read the datacube
        if isinstance(args.cube_module, list) and len(args.cube_module) > 2:
            raise ValueError('Provided cube module must be one or two strings.')
        if isinstance(args.cube_module, str) or len(args.cube_module) == 1:
            UserDataCube = load_object(args.cube_module if isinstance(args.cube_module, str) 
                                       else args.cube_module[0])
        else:
            UserDataCube = load_object(args.cube_module[0], obj=args.cube_module[1])
        #   - Check that the class is derived from DataCube
        if not issubclass(UserDataCube, DataCube):
            raise TypeError('Defined cube object must subclass from mangadap.datacube.DataCube.')

        #   - Import the module used to set the analysis plan
        if isinstance(args.plan_module, list) and len(args.plan_module) > 2:
            raise ValueError('Provided plan module must be one or two strings.')
        if isinstance(args.plan_module, str) or len(args.plan_module) == 1:
            UserPlan = load_object(args.plan_module if isinstance(args.plan_module, str) 
                                       else args.plan_module[0])
        else:
            UserPlan = load_object(args.plan_module[0], obj=args.plan_module[1])
        #   - Check that the class is derived from AnalysisPlan
        if not issubclass(UserPlan, AnalysisPlan):
            raise TypeError('Defined plan object must subclass from '
                            'mangadap.config.analysisplan.AnalysisPlan')

        #   - Instantiate using either the datacube file directly or a
        #     configuration file
        cube = UserDataCube(args.cubefile) if args.config is None \
                    else UserDataCube.from_config(args.config)

        # Read the analysis plan
        plan = UserPlan.default(cube=cube, analysis_path=args.output_path) if args.plan is None \
                    else UserPlan.from_toml(args.plan, cube=cube, analysis_path=args.output_path)

        # Construct the plot for each analysis plan
        for i, key in enumerate(plan.keys()):
            fit_residuals(cube, plan[key], plan.method_path(plan_index=i),
                          plan.method_path(plan_index=i, ref=True),
                          plan.method_path(plan_index=i, qa=True), fwhm=args.beam)

        print('Elapsed time: {0} seconds'.format(time.perf_counter() - t))




#!/usr/bin/env python3

import os
import time
import numpy
import glob

from IPython import embed

from astropy.io import fits
import astropy.constants

#import matplotlib
#matplotlib.use('Agg')
#from matplotlib import pyplot

from mangadap.dapfits import DAPQualityBitMask, DAPMapsBitMask
from mangadap.config import defaults
from mangadap.util.datatable import DataTable
from mangadap.util.fileio import channel_dictionary


class ReferencePropertyTable(DataTable):
    def __init__(self, shape=None):
        datamodel = dict(D4000BIN=dict(typ=float, shape=(2,)),
                         SIGMABIN=dict(typ=float, shape=(2,)),
                         HAEWBIN=dict(typ=float, shape=(2,)),
                         SNRBIN=dict(typ=float, shape=(2,)),
                         NBIN=dict(typ=int, shape=None),
                         PLT=dict(typ=int, shape=None),
                         IFU=dict(typ=int, shape=None),
                         BIN=dict(typ=int, shape=None),
                         MNGTARG1=dict(typ=int, shape=None),
                         MNGTARG3=dict(typ=int, shape=None),
                         Z=dict(typ=float, shape=None),
                         SNR=dict(typ=float, shape=None),
                         STZ=dict(typ=float, shape=None),
                         SIGMA=dict(typ=float, shape=None),
                         HAANR=dict(typ=float, shape=None),
                         HAZ=dict(typ=float, shape=None),
                         HASIGMA=dict(typ=float, shape=None),
                         HAEW=dict(typ=float, shape=None),
                         D4000=dict(typ=float, shape=None))

        keys = list(datamodel.keys())
        super(ReferencePropertyTable,
                self).__init__(keys, [datamodel[k]['typ'] for k in keys],
                               element_shapes=[datamodel[k]['shape'] for k in keys],
                               shape=shape)


def get_maps_files(dapver='3.0.1', daptype='SPX-MILESHC-MASTARHC2'):
#    return ['/uufs/chpc.utah.edu/common/home/sdss/mangawork/manga/spectro/analysis/v3_0_1/{0}/{1}/7443/12704/manga-7443-12704-MAPS-{1}.fits.gz'.format(dapver, daptype), '/uufs/chpc.utah.edu/common/home/sdss/mangawork/manga/spectro/analysis/v3_0_1/{0}/{1}/7443/12701/manga-7443-12701-MAPS-{1}.fits.gz'.format(dapver, daptype), '/uufs/chpc.utah.edu/common/home/sdss/mangawork/manga/spectro/analysis/v3_0_1/{0}/{1}/7443/12703/manga-7443-12703-MAPS-{1}.fits.gz'.format(dapver, daptype), '/uufs/chpc.utah.edu/common/home/sdss/mangawork/manga/spectro/analysis/v3_0_1/{0}/{1}/7443/12705/manga-7443-12705-MAPS-{1}.fits.gz'.format(dapver, daptype), '/uufs/chpc.utah.edu/common/home/sdss/mangawork/manga/spectro/analysis/v3_0_1/{0}/{1}/7443/1901/manga-7443-1901-MAPS-{1}.fits.gz'.format(dapver, daptype), '/uufs/chpc.utah.edu/common/home/sdss/mangawork/manga/spectro/analysis/v3_0_1/{0}/{1}/7443/1902/manga-7443-1902-MAPS-{1}.fits.gz'.format(dapver, daptype), '/uufs/chpc.utah.edu/common/home/sdss/mangawork/manga/spectro/analysis/v3_0_1/{0}/{1}/7443/12702/manga-7443-12702-MAPS-{1}.fits.gz'.format(dapver, daptype)]

    root = os.path.join(defaults.dap_analysis_path(dapver=dapver), daptype)
    return glob.glob(os.path.join(root, '*', '*', '*MAPS*fits.gz'))

# This won't work with hybrid binning scheme!
def include_maps_data(tbl, maps_file, exclude_ancillary=False):
    qual_bm = DAPQualityBitMask()
    bm = DAPMapsBitMask()
   
#    # Targeting bits
#    sdssMaskbits = defaults.sdss_maskbits_file()
#    mngtarg1_bm = BitMask.from_par_file(sdssMaskbits, 'MANGA_TARGET1')
#    mngtarg3_bm = BitMask.from_par_file(sdssMaskbits, 'MANGA_TARGET3')

    with fits.open(maps_file) as hdu:
        # Check if the file was CRITICAL
        if qual_bm.flagged(hdu['PRIMARY'].header['DAPQUAL'], flag='DRPCRIT'):
            return 1
        if exclude_ancillary and hdu[0].header['MNGTARG1'] == 0:
            return 1

        d4000_bins = numpy.unique(tbl['D4000BIN'])
        sigma_bins = numpy.unique(tbl['SIGMABIN'])
        haew_bins = numpy.unique(tbl['HAEWBIN'])
        snr_bins = numpy.unique(tbl['SNRBIN'])
        nsigma = len(sigma_bins)-1
        nd4000 = len(d4000_bins)-1
        nhaew = len(haew_bins)-1
        nsnr = len(snr_bins)-1

        el = channel_dictionary(hdu, 'EMLINE_GFLUX')
        si = channel_dictionary(hdu, 'SPECINDEX')

        plt, ifu = map(lambda x : int(x), os.path.split(maps_file)[1].split('-')[1:3])

        # Input redshift
        z = hdu['PRIMARY'].header['SCINPVEL'] / astropy.constants.c.to('km/s').value

        # Stellar-kinematics binning map
        binid, bin_indx = map(lambda x: x[1:], numpy.unique(hdu['BINID'].data[1,:,:].ravel(),
                                                            return_index=True))

        # Get the D4000, H-alpha EW, and sigma values with S/N greater
        # than one and unmasked
        good_snr = hdu['SPX_SNR'].data.ravel() > 1
        good_haew = numpy.logical_not(bm.flagged(hdu['EMLINE_GEW_MASK'].data[el['Ha-6564']].ravel(),
                                                 flag='DONOTUSE'))
        good_d4000 = numpy.logical_not(bm.flagged(hdu['SPECINDEX_MASK'].data[si['D4000']].ravel(),
                                                  flag='DONOTUSE'))
        good_sigma = numpy.logical_not(bm.flagged(hdu['STELLAR_SIGMA_MASK'].data.ravel(),
                                                  flag='DONOTUSE'))

        indx = (good_snr & good_haew & good_d4000 & good_sigma)[bin_indx]
        snr = hdu['SPX_SNR'].data.ravel()[bin_indx][indx]
        stvel = hdu['STELLAR_VEL'].data.ravel()[bin_indx][indx]
        sigma = hdu['STELLAR_SIGMA'].data.ravel()[bin_indx][indx]
        haanr = hdu['EMLINE_GANR'].data[el['Ha-6564']].ravel()[bin_indx][indx]
        havel = hdu['EMLINE_GVEL'].data[el['Ha-6564']].ravel()[bin_indx][indx]
        hasig = hdu['EMLINE_GSIGMA'].data[el['Ha-6564']].ravel()[bin_indx][indx]
        haew = hdu['EMLINE_GEW'].data[el['Ha-6564']].ravel()[bin_indx][indx]
        d4000 = hdu['SPECINDEX'].data[si['D4000']].ravel()[bin_indx][indx]
        binid = binid[indx]

        d4000_i = numpy.digitize(d4000, d4000_bins[1:-1])
        sigma_j = numpy.digitize(sigma, sigma_bins[1:-1])
        haew_k = numpy.digitize(haew, haew_bins[1:-1])
        snr_l = numpy.digitize(snr, snr_bins[1:-1])

        for i in range(nd4000):
            dindx = d4000_i == i
            for j in range(nsigma):
                sindx = sigma_j == j
                for k in range(nhaew):
                    hindx = haew_k == k
                    for l in range(nsnr):
                        nindx = snr_l == l

                        indx = dindx & sindx & hindx & nindx

                        nbin = numpy.sum(indx)
                        if nbin == 0:
                            continue
                       
                        ii = i*nsigma*nhaew*nsnr + j*nhaew*nsnr + k*nsnr + l
                        tbl['NBIN'][ii] += nbin

                        _snr = snr[indx]
                        # Still find the highest S/N spectrum in each
                        # bin
                        m = numpy.argsort(_snr)[-1]
                        if _snr[m] < tbl['SNR'][ii]:
                            continue

                        tbl['PLT'][ii] = plt
                        tbl['IFU'][ii] = ifu
                        tbl['BIN'][ii] = binid[indx][m]
                        tbl['MNGTARG1'][ii] = int(hdu[0].header['MNGTARG1'])
                        tbl['MNGTARG3'][ii] = int(hdu[0].header['MNGTARG3'])
                        tbl['Z'][ii] = z
                        tbl['SNR'][ii] = snr[indx][m]
                        tbl['STZ'][ii] \
                            = stvel[indx][m]*(1+z)/astropy.constants.c.to('km/s').value + z
                        tbl['SIGMA'][ii] = sigma[indx][m]
                        tbl['HAANR'][ii] = haanr[indx][m]
                        tbl['HAZ'][ii] \
                            = havel[indx][m]*(1+z)/astropy.constants.c.to('km/s').value + z
                        tbl['HASIGMA'][ii] = hasig[indx][m]
                        tbl['HAEW'][ii] = haew[indx][m]
                        tbl['D4000'][ii] = d4000[indx][m]

    return 0
    

def main():
    ofile = 'representative_spectra_selection_v2.fits'
    force = False
    if os.path.isfile(ofile) and not force:
        print('File already exists.')
        return

    exclude_ancillary = True
    daptype = 'SPX-MILESHC-MASTARHC2'
    maps_files = get_maps_files()
    nmaps = len(maps_files)
    print('Found {0} MAPS files.'.format(nmaps))

    # Initialize the output database
    # Sigmas are *uncorrected*
    d4000_bins = numpy.array([ 1.0, 1.2, 1.4, 1.5, 1.6, 1.8, 2.0, 2.2, 2.4 ])
    sigma_bins = numpy.array([0, 25, 50, 75, 100, 150, 200, 250, 2000])
    haew_bins = numpy.array([-1,2,8,16,32,128])
    snr_bins = numpy.array([1,2,5,10,20,40,500])
    nsigma = len(sigma_bins)-1
    nd4000 = len(d4000_bins)-1
    nhaew = len(haew_bins)-1
    nsnr = len(snr_bins)-1

    print(nd4000*nsigma*nhaew*nsnr)
    tbl = ReferencePropertyTable(shape=(nd4000*nsigma*nhaew*nsnr,))

    # Ensure the number in the bin is initialized to 0
    tbl['NBIN'] = 0

    # Set the bin limits
    tbl['D4000BIN'][:,0], tbl['SIGMABIN'][:,0], tbl['HAEWBIN'][:,0], tbl['SNRBIN'][:,0] \
            = tuple( map(lambda x: x.ravel(), numpy.meshgrid(d4000_bins[:-1], sigma_bins[:-1],
                                                             haew_bins[:-1], snr_bins[:-1],
                                                             indexing='ij')))
    tbl['D4000BIN'][:,1], tbl['SIGMABIN'][:,1], tbl['HAEWBIN'][:,1], tbl['SNRBIN'][:,1] \
            = tuple( map(lambda x: x.ravel(), numpy.meshgrid(d4000_bins[1:], sigma_bins[1:],
                                                             haew_bins[1:], snr_bins[1:],
                                                             indexing='ij')))

    # Iterate through the maps files and add the data
    nskipped = 0
    for i in range(nmaps):
        print('{0}/{1}'.format(i+1,nmaps), end='\r')
        nskipped += include_maps_data(tbl, maps_files[i], exclude_ancillary=exclude_ancillary)
    print('{0}/{0}'.format(nmaps))
    print('Number of skipped datacubes: {0}/{1}'.format(nskipped, nmaps))

    # Write the output
    print('Writing: {0}'.format(ofile))
    fits.HDUList([fits.PrimaryHDU(), tbl.to_hdu(name='PAR')]).writeto(ofile, overwrite=True)

#-----------------------------------------------------------------------------

if __name__ == '__main__':
    t = time.perf_counter()
    main()
    print('Elapsed time: {0} seconds'.format(time.perf_counter() - t))






from IPython import embed

import numpy

from mangadap.util.sampling import spectrum_velocity_scale
from mangadap.par.emissionlinedb import EmissionLineDB
from mangadap.proc.emissionlinetemplates import EmissionLineTemplates


def test_init_all():
    wave = numpy.logspace(*numpy.log10([3600., 10000.]), 4563)
    velscale = spectrum_velocity_scale(wave)

    dbs = EmissionLineDB.available_databases()
    for key in dbs.keys():
        if 'MSK' in key:
            continue
        emldb = EmissionLineDB.from_key(key)
        etpl = EmissionLineTemplates(wave, velscale, emldb=emldb)





r"""
Eric EMSELLEM - ESO / CRAL
eric.emsellem@eso.org

.. warning::

    This module is significantly out of date with the rest of the MaNGA DAP
    code, as it was never directly incorporated in the workflow.  It is included
    here for reference.

The main goal of this script is to derive :math:`\lambda_R` and
:math:`V/\sigma` radial profiles, given a set of coordinates
:math:`(x,y)` and kinematic measurements (:math:`V`, :math:`\sigma`, but
also associated flux, :math:`F`)

The standard formulae for :math:`\lambda_R` comes from Emsellem et al.
2007, MNRAS 379, 401 and Emsellem et al. 2011, MNRAS 414, 888, while the
formulae for :math:`V/\sigma` done in the proper way with IFU come from
Binney 2005, 363, 937.

.. math::

    \lambda_R = \frac{\sum F\ |V|}{\sum F\ (V^2 + \sigma^2)^{1/2}}

and

.. math::

    V / \sigma = \left( \frac{\sum F\ V^2}{\sum F\ \sigma^2}
    \right)^{1/2}

The main difference between :math:`\lambda_R` and :math:`V/\sigma` is
that :math:`\lambda_R` is providing a radius-biased view of the
dynamical status and can be used as a proxy for apparent specific
stellar angular momentum, while :math:`V/\sigma` is flux-biased and more
prone to changes due to small central structures.  There is, however, a
nice relation between the two (see the above- mentioned papers) which
holds for regular kinematics.

Both quantities are derived within a given aperture (selected set of
spaxels) and are therefore "cumulative" in the sense that they should
include all spaxels within a certain "radius".  The advised selection is
to follow the moment ellipticity of the system: the derivation of
:math:`\lambda_R` and :math:`V/\sigma` thus require values for the
ellipticity (Eps; :math:`\epsilon`) and position angle (PA;
:math:`\phi_0`).

*Source location*:
    $MANGADAP_DIR/python/mangadap/contrib/LambdaR_2D_forMaNGA.py

*Python2/3 compliance*::

    from __future__ import division
    from __future__ import print_function
    from __future__ import absolute_import
    
    import sys
    if sys.version > '3':
        long = int
        xrange = range

*Imports*::

    import numpy as np
    from numpy import cos, sin, sum, sqrt
    import scipy
    from scipy import interpolate
    import copy

*Revision history*:

    | version 1.1.1 - June 15, 2015: K. Westfall - inclusion in MaNGA
        DAP, minor modifications to docstrings, python 2/3 compliance.
    | version 1.1.0 - March 30, 2015: transformed for MaNGA usage
    | version 1.0.3 - August 20, 2014: added SigmaR
    | version 1.0.2 - August 19, 2014
    | version 1.0.1 - April 23, 2014
    | version 1.0.0 - August 12, 2011 : creation after 2006 module

.. _scipy.interpolate.interp1d: http://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.interp1d.html#scipy.interpolate.interp1d

"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import sys
if sys.version > '3':
    long = int
    xrange = range

import numpy as np
from numpy import cos, sin, sum, sqrt
import scipy
from scipy import interpolate
import copy

def dist2(x1,y1,x2,y2, scale=1.0) :
    return ((x1-x2)**2 + (y1-y2)**2)/scale**2

def guess_regular_grid(xnodes, ynodes, pixelsize=None) :
    """
    Return a regular grid guessed on an irregular one (Voronoi).

    Args:
        xnodes (numpy.array): arrays of Voronoi bin x positions
        ynodes (numpy.array): arrays of Voronoi bin y positions

    Return:
        numpy.ndarray : (xunb, yunb) regular grid for x and y (unbinned)

    """
    ## First deriving a pixel size
    xn_rav, yn_rav = xnodes.ravel(), ynodes.ravel()
    if pixelsize is None :
        pixelsize = derive_pixelsize(xnodes, ynodes)
    minxn = np.int(np.min(xn_rav) / pixelsize) * pixelsize
    minyn = np.int(np.min(yn_rav) / pixelsize) * pixelsize
    xunb, yunb = np.meshgrid(np.arange(minxn, np.max(xn_rav)+pixelsize, pixelsize),
                           np.arange(minyn, np.max(yn_rav)+pixelsize, pixelsize))

    return xunb, yunb

def derive_unbinned_field(xnodes, ynodes, data, xunb=None, yunb=None, mask_array=None) :
    """
    Provide an array of the same shape as the input xunb, and yunb with
    the values derived from the Voronoi binned data

    Args:
        xnodes (numpy.array): arrays of Voronoi bin x positions
        ynodes (numpy.array): arrays of Voronoi bin y positions
        data (numpy.array): values for each node
        xunb (numpy.array): x coordinates of the unbinned data if not
            provided (default) they will be guessed from the nodes
        yunb (numpy.array): y coordinates of the unbinned data if not
            provided (default) they will be guessed from the nodes
        mask_array (numpy.ndarray): array with the same shape than xunb
            providing mask values for positions to ignore

    Returns:
        numpy.ndarray : (xunb, yunb, unbinned_data) arrays with the same
            shape as xunb,

    """
    if xunb is None :
        xunb, yunb = guess_regular_grid(xnodes, ynodes)

    x_rav, y_rav = xunb.ravel(), yunb.ravel()
    xnodes_rav, ynodes_rav = xnodes.ravel(), ynodes.ravel()
    data_rav = data.ravel()
    unbinned_data = np.zeros_like(x_rav, dtype=data_rav.dtype)
    for i in xrange(len(x_rav)) :
        indclosestBin = np.argmin(dist2(x_rav[i], y_rav[i], xnodes_rav, ynodes_rav))
        unbinned_data[i] = data_rav[indclosestBin]

    return xunb, yunb, unbinned_data.reshape(xunb.shape)

###################################################################
# Derive V / Sigma and LambdaR profiles from a set 
# of X, Y, I, V and S
# =================================================================
def Derive_LR_VS_Profiles(X, Y, F, V, S, Rprofile=None, R_EpsPA=None, Eps=None, PA=None,
                          maskDic=None, maskName=None, verbose=False, **kwargs) :
    r"""
    Compute the :math:`\lambda_R` and :math:`V/\sigma` parameters from a
    set of :math:`(x,y)` (positions) and :math:`V`, :math:`\sigma`
    (stellar kinematics).

    If an ellipticity profile is provided, it is used to select the
    points within growing apertures.  Otherwise it uses a default
    ellipse value, constant over the full field.

    Same with the position angle (in degrees), which can vary with
    radius or be constant.

    Rprofile and R_EpsPA are the effective (Geometric) radius.

    .. todo::

        - Documentation out of date.  E.g., not all kwargs listed.
        - Convert result to a true class
        - raise exceptions instead of writing errors and returning
    
    Args:
        X, Y (numpy.ndarray) : Cartesian coordinates. Units can be
            arbitrary as :math:`\lambda_R` and :math:`V/\sigma` are
            dimensionless (but should be linear, so not RA, DEC) *X*,
            and *Y* should be associated with square spaxels but do not
            need to fill in a full rectangular grid.  No default. 
        F, V, S (numpy.ndarray): flux, stellar velocity and velocity
            dispersion measured at *X*, *Y* positions.  Units are
            arbitrary as for *X*, *Y* (since the final result will be
            dimensionless) No default.
        Rprofile (numpy.ndarray): (Optional) Radius sampling imposed by
            the user.  These are the radii where the profiles will be
            derived. Default to None, means that Rsampling will be
            automatically set up.
        R_EpsPA, Eps, PA (numpy.ndarray): ellipticity and Position Angle
            profiles, given as radii (R_EpsPA), ellipticity values (Eps)
            and position angle (degrees). If given (default to None),
            this profile will be used to derive the apertures.
        maskDic (?): (Optional) ?
        maskName (?): (Optional) ?
        verbose (bool): (Optional) Print some additional information if
            relevant.
        Symmetrize (bool): (via kwargs: default is True) to specify
            whether or not the fields should be symmetrized.
        Re (float): (via kwargs: default is 1) effective radius (in
            units of input X, Y)
        Estimate_Vsys (bool): (via kwargs: default is True) If True,
            estimate the systemic velocity (Systemic_Velocity is then
            overwritten). If False, value of Systemic_Velocity will be
            used.
        Systemic_Velocity (float): (via kwargs: default is 0.) Systemic
            Velocity to subtract from V.
        Vaperture (float): (via kwargs: default is 3.) Radius (in units
            of input X, Y) within which V will be summed to estimate the
            systemic velocity.
        Maximum_Velocity (float): (via kwargs: default is 400.)
            Threshold to select out stellar velocities.
        Maximum_Dispersion (float): (via kwargs: default is 500.)
            Threshold to select out stellar velocity dispersion.
        Threshold_Cover (float): (via kwargs: default is 1.25) Fraction
            of pixels which should be covered within an ellipse. Default
            to 1.25, meaning that if more than values for more than 20%
            of the pixels within the ellipse are missing, we stop the
            calculation (as it means that there are not enough points).
        Threshold_Radius (float): (via kwargs: default is 1.15) Same as
            Threshold_Cover but this time in terms of the ratio between
            the actual effective radius, and the effective radius of the
            corresponding ellipse. 
        Min_Radius (float): (via kwargs: default is 1) Minimum radius to
            be considered in the profile.
        Max_Radius (float): (via kwargs: default is 50) Maximum radius
            to be considered in the profile.
        N_Radius (int): (via kwargs: default is 100) Number of points
            within the radius range.

    Returns:
        Result : A class that contains the results of the computation.
            May return None if errors occur.

    """

    ## Defining the Result Class ================================
    class Result :
       def __init__(self):
          self.type = "LambdaR Computation"
    ## ===========================================================

    ## Here is the result structure which will be filled in at the end
    result = Result()
    ## ===========================================================

    ## Default setup =============================================
    result.Symmetrize = kwargs.get('Symmetrize', True)
    result.ExtraCoverageCriterion = kwargs.get('ExtraCoverageCriterion', False)

    result.Re = np.float32(kwargs.get('Re', 1.))

    result.Estimate_Vsys = kwargs.get('Estimate_Vsys', True)
    result.Systemic_Velocity = np.int(kwargs.get('Systemic_Velocity', 0.))
    result.Vaperture = np.float32(kwargs.get('Vaperture', 3.))

    result.Maximum_Velocity = np.float32(kwargs.get('Maximum_Velocity', 500.))
    result.Maximum_Dispersion = np.float32(kwargs.get('Maximum_Dispersion', 500.))

    result.Threshold_Cover = np.float32(kwargs.get('Threshold_Cover', 1.25))
    result.Threshold_Radius = np.float32(kwargs.get('Threshold_Radius', 1.15))

    result.Min_Radius = np.float32(kwargs.get('Min_Radius', 1.))
    result.Max_Radius = np.float32(kwargs.get('Max_Radius', 50.))
    result.N_Radius = np.int(kwargs.get('N_Radius', 100))
    ## ===========================================================

    ## Checking array ============================================
    ref_ShapeX = X.shape
    Shapes = Y.shape, F.shape, V.shape, S.shape
    if not all(s == ref_ShapeX for s in Shapes) :
        print('ERROR: input arrays (X, Y, F, V, S) are not all of the same size')
        print('ERROR: please check the shapes/sizes of these data.')
        return

    if R_EpsPA != None :
        ref_ShapeEps = len(R_EpsPA)
        Shapes = len(Eps), len(PA)
        if not all(s == ref_ShapeEps for s in Shapes) :
            print('ERROR: input arrays (REps, Eps, PA) are not all of the same size')
            print('ERROR: please check the shapes/sizes of these data.')
            return
    ## ===========================================================

    ## Select the right spaxels
    ## All pixels with positive flux, all pixels with positive velocity and dispersion, 
    ## all pixels with velocities lower (amplitude) than the set up Maximum velocity
    ## all pixels with dispersion lower than the set up Maximum dispersion

    ## selFlux is the frst selection to just get an idea of the systemic V
    First_selFlux = (F > 0.) & (S >= 0.) & (np.abs(S) < result.Maximum_Dispersion)
    ## Finally getting the min and max of the velocity and the systemic value from an aperture median
    AperV, minV, maxV = find_centerodd(X[First_selFlux],Y[First_selFlux],V[First_selFlux], Radius=result.Vaperture)

    ## selFlux is the final selection for the spaxels
    selFlux = (F > 0.) & (S >= 0.) & (np.abs(V - AperV) < result.Maximum_Velocity) & (np.abs(S) < result.Maximum_Dispersion)
    ## We combine this with the Masks which may be defined for that galaxy
    mask = select_spaxels(maskDic, maskName, X, Y) * selFlux

    ## Now reducing the arrays to the selected ones
    ## These are the arrays we will use for the calculation
    sX = np.compress(mask, X, axis=0)
    sY = np.compress(mask, Y, axis=0)
    sF = np.compress(mask, F, axis=0)
    sV = np.compress(mask, V, axis=0)
    sS = np.compress(mask, S, axis=0)

    ## Getting the central values and normalising V with Vsys
    ## Finally getting the min and max of the velocity and the systemic value from an aperture median
    AperV, minV, maxV = find_centerodd(sX, sY, sV, Radius=result.Vaperture)
    ## Same for the dispersion (just min and max values)
    minS, maxS = find_centereven(sX,sY,sS)

    if result.Estimate_Vsys : 
        ## Use the estimate aperture value
        result.Systemic_Velocity = AperV
        if verbose:
            print('Central velocity measured is : {0}'.format(AperV))
    else : 
        if verbose:
            print('Systemic Velocity value use is : {0}'.format(result.Systemic_Velocity))

    # Return the symmetrised points if Symmetrize is true
    if result.Symmetrize :
        if verbose : 
            print('Symmetrizing the points')
        nsV = SymmetrizeField(sX, sY, sV - result.Systemic_Velocity, odd=1)
        nsS = SymmetrizeField(sX, sY, sS, odd=0)
    else :
        nsV = sV - result.Systemic_Velocity
        nsS = sS

    ## Define the Radii to be used for the profile
    ## If not provided by Rprofile, then build one using Min/Max_Radius and N_Radius
    if Rprofile == None :
       Rprofile = np.linspace(result.Min_Radius, result.Max_Radius, result.N_Radius)

    ## Creation of a big grid of step
    dist = sqrt((X - X[0])**2 + (Y - Y[0])**2)
    ## looking for the minimum radius (which is not zero) which should be the Size of the smallest Spaxel
    binstep = np.min(dist[dist>0])
    if verbose: 
        print('BINSTEP is {0}'.format(binstep))

    ## Computing the Map Grid Corners (after masking)
    X1, X2, Y1, Y2 = Get_CornersArray(sX, sY, binstep)

    ## And computing the large rectangular grid containing the full set of spaxels
    Nspaxels = np.max([X.min()/binstep, X.max()/binstep, Y.min()/binstep, Y.max()/binstep])
    XYsample = np.arange(-Nspaxels*binstep, (Nspaxels+1)*binstep, binstep)
    GX, GY = np.meshgrid(XYsample, XYsample)
    ## Now computing the Full grid corners
    GX1, GX2, GY1, GY2 = Get_CornersArray(GX, GY, binstep)

    ## Define the ellipticity profiles to be used ================
    if R_EpsPA == None :
        if (Eps == None)  :
            print('Error: no Ellipticity provided')
            return 
        elif (PA == None) :
            print('Error: no Position Angle provided')
            return 
        else :
            R_EpsPA = Rprofile
            Eps = np.zeros_like(Rprofile) + Eps
            PA = np.zeros_like(Rprofile) + PA

    ## Defining the function to interpolate Eps and PA
    profEps = interpolate.interp1d(R_EpsPA, Eps, kind="linear")
    profPA = interpolate.interp1d(R_EpsPA, PA, kind="linear")
    ## ===========================================================

    ## Contracting Rprofile with the input R_EpsPA ==================
    compactRprofile = Rprofile[(Rprofile > np.min(R_EpsPA)) & (Rprofile < np.max(R_EpsPA))]

    ## Start the Loop ============================================
    Npoints = len(compactRprofile)

    LambdaR = np.zeros(Npoints, dtype=np.float32)
    FlagRadius = np.ones(Npoints, dtype=np.int32)

    VS = np.zeros_like(LambdaR)
    sumF = np.zeros_like(LambdaR)
    FV2 = np.zeros_like(LambdaR)
    FS2 = np.zeros_like(LambdaR)
    FMU2 = np.zeros_like(LambdaR)
    RFV = np.zeros_like(LambdaR)
    RFMU = np.zeros_like(LambdaR)
    SemiMajorRadius = np.zeros_like(LambdaR)
    EffRadius = np.zeros_like(LambdaR)
    GEffRadius = np.zeros_like(LambdaR)

    ## Starting the actual loop on the profile radii
    for i in range(Npoints) :
       locRadius = compactRprofile[i]    ## This is the geometric radius
       Eps_Here = profEps(locRadius)     ## THe ellipticity at that radius
       if Eps_Here >= 1. :
          print('Error: we get an ellipticity of 1 at radius {0}'.format(locRadius))
          continue

       PA_Here = profPA(locRadius)       ## The PA at that radius
       SemiMajorRadius[i] = locRadius / sqrt(1. - Eps_Here)  ## Semi major axis radius

       ## Rotating the axis to have it aligned
       rsX, rsY = rotPA(sX,sY, PA_Here)
       rellipse = XY_toSemiMajor(rsX, rsY, Eps_Here) ## SEMI MAJOR AXIS

       ## Rotating the Grid in the same way
       rGX, rGY = rotPA(GX,GY, PA_Here)
       rGellipse = XY_toSemiMajor(rGX.ravel(), rGY.ravel(), Eps_Here) ## SEMI MAJOR AXIS

       ## Selection for the present Radius
       selection = (rellipse <= SemiMajorRadius[i]) 
       Gselection = (rGellipse <= SemiMajorRadius[i]) 

       ## Rotating the Corners
       rXa, rXb, rXc, rXd, rYa, rYb, rYc, rYd = Rotate_Corners(X1, X2, Y1, Y2, PA_Here)
       ## Deriving the corresponding Major-Axis radius
       rElla = XY_toSemiMajor(rXa, rYa, Eps_Here) 
       rEllb = XY_toSemiMajor(rXb, rYb, Eps_Here) 
       rEllc = XY_toSemiMajor(rXc, rYc, Eps_Here) 
       rElld = XY_toSemiMajor(rXd, rYd, Eps_Here) 
       ## Also just the radius
       ra = XY_toSemiMajor(rXa, rYa, 0.0) 
       rb = XY_toSemiMajor(rXb, rYb, 0.0) 
       rc = XY_toSemiMajor(rXc, rYc, 0.0) 
       rd = XY_toSemiMajor(rXd, rYd, 0.0) 

       ## Same with the Full Grid
       rGXa, rGXb, rGXc, rGXd, rGYa, rGYb, rGYc, rGYd = Rotate_Corners(GX1, GX2, GY1, GY2, PA_Here)
       rGElla = XY_toSemiMajor(rGXa, rGYa, Eps_Here) 
       rGEllb = XY_toSemiMajor(rGXb, rGYb, Eps_Here) 
       rGEllc = XY_toSemiMajor(rGXc, rGYc, Eps_Here) 
       rGElld = XY_toSemiMajor(rGXd, rGYd, Eps_Here) 

       Allr = np.concatenate((ra,rb,rc,rd))
       AllEllr = np.concatenate((rElla,rEllb,rEllc,rElld))
       AllGEllr = np.concatenate((rGElla,rGEllb,rGEllc,rGElld))
       MaxR = np.max(Allr)

       ## Selecting the coordinates within this region
       srsX = np.compress(selection, rsX, axis=0)
       srsY = np.compress(selection, rsY, axis=0)
       ssX = np.compress(selection, sX, axis=0) 
       ssY = np.compress(selection, sY, axis=0) 
       ssR = sqrt(ssX**2 + ssY**2)
       ssF = np.compress(selection, sF, axis=0)
       ssV = np.compress(selection, nsV, axis=0)
       ssS = np.compress(selection, nsS, axis=0)

       ## Calculate the full Area on the full Grid
       ssGX = np.compress(Gselection, GX.ravel(), axis=0)
       ## The effective radius is then just the SQRT of the AREA
       GEffRadius[i] = np.sqrt(len(ssGX) * binstep * binstep / np.pi)

       ## Calculate the effective radius for the selection on Input Data
       EffRadius[i] = sqrt(len(ssF) * binstep * binstep / np.pi)

       ## For V/S calculate I * V^2 and I * S^2 as in Binney 2005
       FV2[i] = sum(ssF * ssV**2, axis=0)
       FS2[i] = sum(ssF * ssS**2, axis=0)

       ## For Sigma_e
       FMU2[i] = FV2[i] + FS2[i]
       sumF[i] = sum(ssF)

       ## For LambdaR, derive R * I * ABS(V)
       ## and R * I * (V^2 +  S^2)
       RFV[i] = sum(ssF * np.abs(ssV) * ssR, axis=0)
       RFMU[i] = sum(ssF * sqrt(ssV**2 + ssS**2) * ssR, axis=0)

       #  Lp, Lm, momEps, momPA = Flux_Inertia(ssX, ssY, ssF)

       ## Check if the Effective radius passes the Threshold
       ## If not, Flag the point to 0
       if (GEffRadius[i] / EffRadius[i] > result.Threshold_Radius) :
          FlagRadius[i] = 0.

       ## Now check other criteria by looking at how many pixels are intersecting
       ## This criterion should in principle be ignored
       if result.ExtraCoverageCriterion :
          nPixEll = len(sX[check_cross_pixel(AllEllr,SemiMajorRadius[i])])
          nPixGEll = len(GX[check_cross_pixel(AllGEllr,SemiMajorRadius[i])])
          if nPixEll == 0 :
             FlagRadius[i] = 0
          elif (((nGPixEll*1.00)/(nPixEll*1.00)) > result.Threshold_Cover) :
             FlagRadius[i] = 0

    ## Finalising the calculation
    ## V / Sigma
    VS = sqrt(FV2 / FS2)
    ## LambdaR
    LambdaR = RFV / RFMU
    ## Aperture Sigma
    SigmaR = sqrt(FMU2 / sumF)

    ## Interpolation at Various Radii : Re/2, Re 
    selRadius = (FlagRadius == 1)
    EffRadiusN = EffRadius / result.Re
    MaxEffRadius = np.max(EffRadius[selRadius])
    MaxEffRadiusN = MaxEffRadius / result.Re

    ## Calculation at 0.5, 1 and 2 Re
    RadiusN_for_re = np.minimum(1.0,  MaxEffRadiusN)
    RadiusN_for_2re = np.minimum(2.0,  MaxEffRadiusN)
    RadiusN_for_re2 = np.minimum(0.5,  MaxEffRadiusN)

    VSre2 = interp_value_fromR(EffRadiusN, VS, RadiusN_for_re2)
    LambdaRre2 = interp_value_fromR(EffRadiusN, LambdaR, RadiusN_for_re2)
    Sigmare2 = interp_value_fromR(EffRadiusN, SigmaR, RadiusN_for_re2)
    VSre = interp_value_fromR(EffRadiusN, VS, RadiusN_for_re)
    LambdaRre = interp_value_fromR(EffRadiusN, LambdaR, RadiusN_for_re)
    Sigmare = interp_value_fromR(EffRadiusN, SigmaR, RadiusN_for_re)
    VS2re = interp_value_fromR(EffRadiusN, VS, RadiusN_for_2re)
    LambdaR2re = interp_value_fromR(EffRadiusN, LambdaR, RadiusN_for_2re)
    Sigma2re = interp_value_fromR(EffRadiusN, SigmaR, RadiusN_for_2re)

    ## Just adding all the relevant arrays into the structure
    result.LambdaR = LambdaR
    result.VS = VS
    result.GEffRadius = GEffRadius
    result.EffRadius = EffRadius
    result.EffRadiusN = EffRadiusN
    result.FlagRadius = FlagRadius
    result.MaxEffRadius = MaxEffRadius
    result.LambdaR_re = LambdaRre
    result.LambdaR_re2 = LambdaRre2
    result.LambdaR_2re = LambdaR2re
    result.RadiusN_for_re = RadiusN_for_re
    result.RadiusN_for_2re = RadiusN_for_2re
    result.RadiusN_for_re2 = RadiusN_for_re2
    result.Sigma_re = Sigmare
    result.Sigma_re2 = Sigmare2
    result.Sigma_2re = Sigma2re
    result.VS_re = VSre
    result.VS_re2 = VSre2
    result.VS_2re = VS2re
    result.binstep = binstep
    result.R_EpsPA = R_EpsPA
    result.Eps = Eps
    result.PA = PA
    result.SemiMajorRadius = SemiMajorRadius
    return result
# ===============================================================
###################################################################
def interp_value_fromR(R=None, val=None, radius=None) :
    """
    Interpolation of a profile at a certain radius.  Check if the radius
    is really within the given input range

    If radius out of range, return the closest (in radius) value. This
    is thus different from the option of bounds_error of
    `scipy.interpolate.interp1d`_ which is used here.

    Args:
        R (numpy.ndarray): (Optional) Input radii
        val (numpy.ndarray): (Optional) Input values. Should have the
            same size as R.
        radius (float): (Optional) Radius at which to interpolate

    Returns:
        float : Value interpolated from the input data.

    """
    if (radius < R[0]) :
        return val[0]

    elif (radius > R[-1]) :
        return val[-1]

    else :
        fval = interpolate.interp1d(R, val, kind="linear")
        return fval([radius])[0]
# =================================================================
####################################################################
# def Flux_Inertia(X, Y, Flux) :
#     """
#     Derive the moment of inertia from a flux map
#     Given X, Y and Flux values
# 
#     Return minor, major, ellipticity, and PA in degrees
#     """
#     momI = sum(Flux, axis=0)
#     if momI == 0. :
#         return 0., 0., 1., 0.
#     momIX = sum(Flux * X, axis=0) / momI
#     momIY = sum(Flux * Y, axis=0) / momI
#     a = sum(Flux * X * X, axis=0) / momI - momIX**2
#     b = sum(Flux * Y * Y, axis=0) / momI - momIY**2
#     c = sum(Flux * X * Y, axis=0) / momI - momIX * momIY
#     if c == 0 :
#         if a == 0. :
#             return b, a, 0., 0.
#         if b > a :
#             return b, a, 1. - sqrt(a / b), 0.
#         else :
#             if b == 0.:
#                 return a,b,0.,0.
#             else :
#                 return a, b, 1. - sqrt(b/a), 90.
# 
#     delta = (a - b)**2. + 4 * c * c 
#     Lp2 = ((a + b) + sqrt(delta)) / 2.
#     Lm2 = ((a + b) - sqrt(delta)) / 2.
#     Lp = sqrt(np.maximum(Lp2, 0.))
#     Lm = sqrt(np.maximum(Lm2, 0.))
#     eps = (Lp - Lm) / Lp
#     PA = np.rad2deg((np.arctan((b - Lp2) / c))
#     return Lp, Lm, eps, PA
# ==================================================================

####################################################################
# Function to see if the ellipse is basically 
#          in or out or intersects the spaxel
####################################################################
def check_cross_pixel(rpixels, rlimit) :
    """
    Check whether the ellipse is going THROUGH the pixel or not.  Return
    an array of True (intersecting) or False (Not intersecting).
    """
    signC = np.sign(rpixels - rlimit)
    # True means the pixel has the circle intersecting it
    # False means the ellipse is in or out
    return np.abs(signC.sum(axis=0)) < 4.0
# ==================================================================

####################################################################
# Define the central value of a map
# Case of an odd symmetry
def find_centerodd(X, Y, Z, Radius=3.0) :
    """
    Find central value for an odd sided field.

    Args:
        X (numpy.ndarray): x coordinates
        Y (numpy.ndarray): y coordinates
        Z (numpy.ndarray): values
        Radius (float): (Optional) Radius within which to derive the
            central value.

    Returns:
        float: The central value and some amplitude value (range about
            that value?).

    """
    # First select the points which are non-zero and compress
    ravZ = np.ravel(Z)
    sel = (ravZ != 0.)
    selz = np.compress(sel,ravZ)

    ## Select the central points for the central value
    selxy = np.ravel((np.abs(X) < Radius) & (np.abs(Y) < Radius)) & sel
    selzxy = np.compress(selxy, ravZ)
    cval = np.median(selzxy)

    sig = np.std(selz)
    sselz = np.compress(np.abs(selz - cval) < 3 * sig, selz - cval)
    ampl = np.max(np.abs(sselz)) / 1.1
    return cval, cval - ampl, cval + ampl
# ==================================================================

####################################################################
# Define the central value of a map
# Case of an even symmetry
def find_centereven(X, Y, Z, Radius=3.0, sel0=True) :
    """
    Find the central value for an even sided field.

    Args:
        X (numpy.ndarray): x coordinates
        Y (numpy.ndarray): y coordinates
        Z (numpy.ndarray): values
        Radius (float): (Optional) Radius within which to derive the
            central value.
        sel0 (bool): (Optional) ?

    Returns:
        float : The minimum and maximum value.

    """
    ravZ = np.ravel(Z)
    if sel0 : sel = (ravZ != 0.)
    else : sel = (np.abs(ravZ) >= 0.)
    sel = np.ravel((np.abs(X) < Radius) & (np.abs(Y) < Radius)) & sel
    selz = np.compress(sel, ravZ)
    maxz = np.max(selz)
    minz = np.min(selz)

    return minz, maxz
# ==================================================================

########## point symmetrization of the points #####################
# Use a simple symmetry since pixels are centred
# So reading the opposite pixel
###################################################################
def SymmetrizeField(X, Y, Z, odd=1, maskDic=None, maskName=None) :
    """
    Symmetrize a field. Only works when spaxels are located around a
    central spaxel.

    Args:
        X (numpy.ndarray): x coordinates
        Y (numpy.ndarray): y coordinates
        Z (numpy.ndarray): input values
        odd (bool): (Optional) True if it is odd w.r.t. centre
            (X=0,Y=0). False if even.
        maskDic, maskName: (Optional) Dictionary and name of masks

    Returns:
        numpy.ndarray : The symmetrized array (symmetrized Z values)

    """
    newz = copy.copy(Z)

    if odd == 1:
        weight = -1.
    else :
        weight = 1.

    ## Interpolation
    xp = np.zeros(1, np.float32)
    yp = np.zeros(1, np.float32)
    for i in range(len(Z)) :
        ## Look for the symmetric point (w.r.t centre)
        xp[0] = -X[i]
        yp[0] = -Y[i]
        ## Only one spaxel to test
        selmask = select_spaxels(maskDic, maskName, xp, yp) # Mask the bad regions
        ## If the symmetric spaxel is not good then I just use the value within the masked region
        ## Otherwise I average neighbouring spaxels
        if selmask[0] ==  True :
            nz = Z[(abs(X - xp[0]) + abs(Y - yp[0])) < 1.e-3]
            ## If found some spaxels, average the values
            if len(nz) > 0 :
                newz[i] = (nz[0] * weight + Z[i]) / 2.0

    return newz
# ===============================================================
###################################################################
def Get_CornersArray(X, Y, binstep) :
     """
     Return corners given a set of X and Y and a fixed binstep.
     """
     X1, X2 = X + binstep/2., X - binstep/2.
     Y1, Y2 = Y + binstep/2., Y - binstep/2.
     return X1, X2, Y1, Y2
# ===============================================================
####################################################################
## Rotating coordinates, this includes a 90 degrees rot to
## have the major axis along the Y axis transformed in X
## So rotangle is the POSITION ANGLE from Top, counter-clockwise
def rotPA(X,Y, PA) :
     """ Rotation of X and Y
         PA in degrees
     """ 
     PA_rad = np.deg2rad(PA)
     Xr = - X * sin(PA_rad) + Y * cos(PA_rad)
     Yr = - X * cos(PA_rad) - Y * sin(PA_rad)
     return Xr, Yr
# ==================================================================
###################################################################
def Rotate_Corners(X1, X2, Y1, Y2, PA) :
     """
     Rotate corners of a grid and return all corners
     Using a PA in degrees
     """
     rXa, rYa = rotPA(X1, Y1, PA)
     rXb, rYb = rotPA(X2, Y2, PA)
     rXc, rYc = rotPA(X1, Y2, PA)
     rXd, rYd = rotPA(X2, Y1, PA)
     return rXa, rXb, rXc, rXd, rYa, rYb, rYc, rYd
# ===============================================================
###################################################################
def XY_toSemiMajor(X, Y, Eps) :
     """
     Transform X, Y into Semi-Major axis radii using a given ellipticity

     Args:
        X (numpy.ndarray):
            Array of X coordinate
        Y (numpy.ndarray):
            Array of Y coordinate
        Eps (:obj:`float`):
            Ellipticity (1-b/a).

     """
     return sqrt(X**2 + Y**2 / (1. - Eps)**2)
# ===============================================================

####################################################################
# Function to select (Mask) good values from a map
# Using both Rectangle and Circular Zones from the Selection_Zone class
#
# Input is name of galaxy, and coordinates
####################################################################
def select_spaxels(maskDic, maskName, X, Y) :
   ## All spaxels are set to GOOD (True) first
   selgood = (X**2 >= 0)

   ## If no Mask is provided, we just return the full set of input X, Y
   if maskDic == None :
      return selgood

   ## We first check if the maskName is in the list of the defined Masks
   ## If the galaxy is not in the list, then the selection is all True
   if (maskDic.has_key(maskName)) :
      ## The mask is defined, so Get the list of Regions
      ## From the defined dictionary
      listRegions = maskDic[maskName]
      ## For each region, select the good spaxels
      for region in  listRegions :
         selgood = selgood & region.select(X, Y)

   return selgood
#=================================================================
####################################################################
# Parent class for the various types of Zones
####################################################################
class Selection_Zone :
    """
    Parent class for Rectangle_Zone and Circle_Zone
    """
    def __init__(self, params=None) :
        self.params = params

#=================================================================
####################################################################
class Rectangle_Zone(Selection_Zone) :
    type = "Rectangle"
    def select(self, xin, yin) :
        """ Define a selection within a rectangle
            It can be rotated by an angle theta (in degrees) """
        if self.params == None :
           return (xin**2 >=0)
        [x0,y0,length,width,theta] = self.params
        dx = xin - x0
        dy = yin - y0
        thetarad = theta * np.pi / 180.
        nx =   dx * cos(thetarad) + dy * sin(thetarad) 
        ny = - dx * sin(thetarad) + dy * cos(thetarad) 
        selgood = (np.abs(ny) > width/2.) | (np.abs(nx) > length/2.)
        return selgood
#=================================================================
####################################################################
class Circle_Zone(Selection_Zone) :
    type = "Circle"
    def select(self, xin, yin) :
        """ Define a selection within a circle """
        if self.params == None :
           return (xin**2 >=0)
        [x0,y0,radius] = self.params
        selgood = (sqrt((xin - x0)**2 + (yin - y0)**2) > radius)
        return selgood
#=================================================================

####################################################################
## Example of Selection Zone for a set of galaxies
## A3DMask = {'IC0598'      : [Circle_Zone([-13.6, +7.9, 2.6])],
##   'IC0676'    : [Circle_Zone([+10.5, +18.2, 6.2])]
## , 'NGC0448'   : [Circle_Zone([+19.0, -5.6, 2.2])]
## , 'NGC0509'   : [Circle_Zone([-19.1, -8.0, 3.0]), Circle_Zone([-0.9, +11.1, 3.0])]
## , 'NGC0661'   : [Circle_Zone([+11.9, -19.1, 4.2])]
## , 'NGC4486'   : [Rectangle_Zone([0.0,0.0,2.0]),Rectangle_Zone([12.0,4.8,12.0,4.0,23])]
## , 'NGC4486A'  : [Circle_Zone([+1.0, -2.3, 3.5]), Circle_Zone([+1.4, -9.8, 2.4]), Circle_Zone([-4.1, -13.3, 2.4]), Circle_Zone([+7.6, -14.1, 2.4])]
####################################################################




import os
import warnings
import tqdm
import requests
import netrc

from IPython import embed

from mangadap.tests.util import remote_data_file, remote_data_files
from mangadap.tests.util import drp_test_version, dap_test_version

try:
    NETRC = netrc.netrc()
except Exception as e:
    NETRC = None
    warnings.warn('Could not load ~/.netrc file.  Attempting to pull from DR17.')

HOST = 'data.sdss.org'
if NETRC is not None and HOST not in NETRC.hosts:
    NETRC = None
    warnings.warn('Host data.sdss.org is not defined in your ~/.netrc file.  '
                  'Attempting to pull from DR17.')

def download_file(remote_root, usr, passwd, local_root, file, overwrite=False):
    """
    Thanks to 
    https://stackoverflow.com/questions/37573483/progress-bar-while-download-file-over-http-with-requests/37573701
    """
    #Beware of how this is joined!
    url = f'{remote_root}{file}'
    ofile = os.path.join(local_root, file)

    if os.path.isfile(ofile):
        if overwrite:
            warnings.warn(f'Overwriting existing file: {ofile}')
            os.remove(ofile)
        else:
            warnings.warn('{ofile} exists.  To overwrite, set overwrite=True.')
            return

    print(f'Downloading: {url}')
    # Streaming, so we can iterate over the response.
    r = requests.get(url, stream=True, auth=(usr, passwd))
    if r.status_code == 404:
        raise ValueError('Requested URL returned 404')
    # Total size in bytes.
    total_size = int(r.headers.get('content-length', 0))
    block_size = 1024 #1 Kibibyte
    t = tqdm.tqdm(total=total_size, unit='iB', unit_scale=True)
    with open(ofile, 'wb') as f:
        for data in r.iter_content(block_size):
            t.update(len(data))
            f.write(data)
    t.close()
    if total_size != 0 and t.n != total_size:
        raise ValueError('Downloaded file may be corrupted.')


def main():

    overwrite = False
    if NETRC is None:
        usr, acc, passwd = None, None, None
    else:
        usr, acc, passwd = NETRC.authenticators(HOST)

    sas_dir = 'dr17' if NETRC is None else 'mangawork'
    sas_root = f'https://{HOST}/sas/{sas_dir}/manga/spectro'

    version = drp_test_version
    files = remote_data_files()
    plates = [f.split('-')[1] for f in files]

    local_root = remote_data_file()
    if not os.path.isdir(local_root):
        os.makedirs(local_root)

    # Get the spectral data
    for plate, f in zip(plates, files):
        url_root = f'{sas_root}/redux/{drp_test_version}/{plate}/stack/'
        try:
            download_file(url_root, usr, passwd, local_root, f, overwrite=overwrite)
        except Exception as e:
            print(str(e))
            continue

    # Get the DRPComplete file
    f = f'drpcomplete_{drp_test_version}.fits'
    url_root = f'{sas_root}/analysis/{drp_test_version}/{dap_test_version}/common/'
    try:
        download_file(url_root, usr, passwd, local_root, f, overwrite=overwrite)
    except Exception as e:
        print(str(e))

    # Get the DRPall file
    f = f'drpall-{drp_test_version}.fits'
    url_root = f'{sas_root}/redux/{drp_test_version}/'
    try:
        download_file(url_root, usr, passwd, local_root, f, overwrite=overwrite)
    except Exception as e:
        print(str(e))


if __name__ == '__main__':
    main()





"""
Implement the derived class for MaNGA row-stacked spectra.

----

.. include license and copyright
.. include:: ../include/copy.rst

----

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""
import os
import warnings

from IPython import embed

import numpy

from astropy.io import fits

from ..config.manga import MaNGAConfig
from ..util.drpbitmask import DRPFitsBitMask
from ..util.constants import DAPConstants
from ..util.filter import interpolate_masked_vector
from .rowstackedspectra import RowStackedSpectra


class MaNGARSS(MaNGAConfig, RowStackedSpectra):
    r"""
    Container class for MaNGA row-stacked spectra.

    For additional description and attributes, see the two base
    classes.

    See :func:`from_plateifu` to instantiate using the plate and ifu
    numbers. See :func:`from_config` to instantiate from a
    configuration file.

    Args:
        ifile (:obj:`str`):
            The file with the MaNGA row-stacked spectra. The name is
            expected to follow the nominal naming convention, which
            is used to parse the plate, ifudesign, and whether or not
            the spectra are binned logarithmically in wavelength.
        sres_ext (:obj:`str`, optional):
            The extension to use when constructing the spectral
            resolution vectors. See :func:`spectral_resolution`.
        sres_fill (:obj:`bool`, optional):
            Fill masked values by interpolation. Default is to leave
            masked pixels in returned array.
    """
    def __init__(self, ifile, sres_ext=None, sres_fill=True):

        # Use the configuration class method to construct the configuration
        # based on the input data file
        cfg = MaNGAConfig.from_file(ifile)
        # and then use the parent class init to set it to self.  NOTE: This is
        # round about, but the way I got it to work.  There must be a better way
        # to do this...
        MaNGAConfig.__init__(self, cfg.plate, cfg.ifudesign, mode=cfg.mode, log=cfg.log,
                             drpver=cfg.drpver, redux_path=cfg.redux_path,
                             chart_path=cfg.chart_path, directory_path=cfg.directory_path)

        # Define the relevant BitMask object
        bitmask = DRPFitsBitMask(mode='RSS')

        # Open the file and initialize the base class
        with fits.open(ifile) as hdu:
            print('Reading MaNGA row-stacked spectra data ...', end='\r')
            prihdr = hdu[0].header
            fluxhdr = hdu['FLUX'].header
            self.sres_ext, sres = MaNGAConfig.spectral_resolution(hdu, ext=sres_ext, fill=sres_fill)
            self.sres_fill = sres_fill
            sres = sres.filled(0.0)

            # The negative for XPOS is because the data in that
            # extension has negative offsets going toward increasing
            # RA. Opposite of the RowStackedSpectra convention. The
            # area of the fiber aperture is normalized to pi arcsec^2
            # by the DRP.
            RowStackedSpectra.__init__(self, hdu['WAVE'].data, hdu['FLUX'].data,
                                       ivar=hdu['IVAR'].data, mask=hdu['MASK'].data,
                                       bitmask=bitmask, sres=sres, xpos=-hdu['XPOS'].data,
                                       ypos=hdu['YPOS'].data, area=numpy.pi,
                                       log=self.log, prihdr=prihdr, fluxhdr=fluxhdr)
        print('Reading MaNGA row-stacked spectra data ... DONE')

    @classmethod
    def from_plateifu(cls, plate, ifudesign, log=True, drpver=None, redux_path=None,
                      chart_path=None, directory_path=None, **kwargs):
        """
        Construct a :class:`mangadap.datacube.manga.MaNGARSS`
        object based on its plate-ifu designation.

        This is a simple wrapper function that calls
        :func:`default_paths` to construct the file names, and then
        calls the class instantiation method.

        Args:
            plate (:obj:`int`):
                Plate number
            ifudesign (:obj:`int`):
                IFU design
            log (:obj:`bool`, optional):
                Use the spectra that are logarithmically binned in
                wavelength.
            drpver (:obj:`str`, optional):
                DRP version, which is used to define the default DRP
                redux path. Default is defined by
                :func:`mangadap.config.manga.drp_version`
            redux_path (:obj:`str`, optional):
                The path to the top level directory containing the
                DRP output files for a given DRP version. Default is
                defined by
                :func:`mangadap.config.manga.drp_redux_path`.
            chart_path (:obj:`str`, optional):
                The path to the directory containing the finding chart images
                for this plate.  Default is defined by
                :func:`mangadap.config.manga.drp_finding_chart_path`.
            directory_path (:obj:`str`, optional):
                The exact path to the DRP file. Default is defined by
                :func:`mangadap.config.manga.drp_directory_path`.
                Providing this ignores anything provided for
                ``drpver`` or ``redux_path``.
            **kwargs:
                Keyword arguments passed directly to the primary
                instantiation method; see
                :class:`mangadap.spectra.manga.MaNGARSS`.
        """
        # Use the input to instantiate the configuration
        cfg = MaNGAConfig(plate, ifudesign, mode='RSS', log=log, drpver=drpver,
                          redux_path=redux_path, chart_path=chart_path,
                          directory_path=directory_path)
        # Then use the configuration to instantiate the object
        return cls(str(cfg.file_path), **kwargs)

    @classmethod
    def from_config(cls, cfgfile):
        """
        Construct a :class:`mangadap.datacube.manga.MaNGADataCube` or
        :class:`mangadap.spectra.manga.MaNGARSS` object from a
        configuration file.

        Using the data read from the configuration file, the method
        instantiates the class using
        :func:`mangadap.spectra.manga.MaNGARSS.from_plateifu` or
        :func:`mangadap.datacube.manga.MaNGADataCube.from_plateifu`.
        This method will therefore fault for this base class!

        The format of the configuration file is:

        .. todo::

            Fill this in.

        Args:
            cfgfile (:obj:`str`):
                Configuration file
            drpver (:obj:`str`, optional):
                DRP version, which is used to define the default DRP
                redux path. Overrides any value in the configuration
                file.
            redux_path (:obj:`str`, optional):
                The path to the top level directory containing the
                DRP output files for a given DRP version. Overrides
                any value in the configuration file.
            directory_path (:obj:`str`, optional):
                The exact path to the DRP file. Providing this
                ignores anything provided for ``drpver`` or
                ``redux_path``. Overrides any value in the
                configuration file.
        """
        # First use the base class method to instantiate the configuration
        cfg = MaNGAConfig.from_config(cfgfile, mode='RSS')

        # Then read the file (again...) to get the relevant keyword arguments
        full_cfg = DefaultConfig(cfgfile, interpolate=True)
        kwargs = {}
        kwargs['z'] = full_cfg.getfloat('z')
        kwargs['vdisp'] = full_cfg.getfloat('vdisp')
        kwargs['ell'] = full_cfg.getfloat('ell')
        kwargs['pa'] = full_cfg.getfloat('pa')
        kwargs['reff'] = full_cfg.getfloat('reff')
        kwargs['sres_ext'] = full_cfg.get('sres_ext')
        kwargs['sres_fill'] = full_cfg.getbool('sres_fill', default=True)
        kwargs['covar_ext'] = full_cfg.get('covar_ext')
        return cls(str(cfg.file_path), **kwargs)

    def pointing_offset(self):
        """
        Return the offsets in RA and DEC between the pointing
        coordinates (``IFURA``, ``IFUDEC``) and the designated object
        center coordinates (``OBJRA``, ``OBJDEC``), drawn from the
        primary header of the DRP fits file.

        Returns:
            :obj:`float`: Sky-right arcsecond offsets in RA and DEC.
        """
        return ((self.prihdr['IFURA'] - self.prihdr['OBJRA']) \
                    * numpy.cos(numpy.radians(self.prihdr['OBJDEC'])) * 3600.), \
               ((self.prihdr['IFUDEC'] - self.prihdr['OBJDEC']) * 3600.)

    def mean_sky_coordinates(self, offset=True, **kwargs):
        r"""
        Compute the weighted or unweighted mean sky coordinates for
        each spectrum.

        This is a simple wrapper for
        :func:`mangadap.spectra.rowstackedspectra.RowStackedSpectra.mean_sky_coordinates`;
        see that method for additional keyword arguments.

        .. warning::

            Flux-weighting the coordinates can produce spurious
            results in low-flux regimes.

        Args:
            offset (:obj:`bool`, optional):
                Offset the coordinates in :attr:`xpos` and
                :attr:`ypos` by the difference between the IFU and
                OBJ coordinates in the header (see
                :func:`pointing_offset`). The coordinates are
                nominally defined with respect to the IFU pointing
                center, such that this offsets the coordinates to be
                centered on the galaxy.
            **kwargs:
                Passed directly to
                :func:`mangadap.spectra.rowstackedspectra.RowStackedSpectra.mean_sky_coordinates`;

        Returns:
            :obj:`tuple`: Two `numpy.ndarray`_ objects with the
            fiducial x and y on-sky positions of each spectrum in
            arcseconds relative to a given center. The shape of the
            two arrays is :math:`(N_{\rm spec},)`.
        """
        # TODO: offset cannot be a keyword in kwargs.
        _offset = self.pointing_offset() if offset else None
        return super(MaNGARSS, self).mean_sky_coordinates(offset=_offset, **kwargs)

    def binned_on_sky_area(self, bin_indx, offset=True, **kwargs):
        r"""
        Compute the on-sky area of a set of binned spectra.

        This is a simple wrapper for
        :func:`mangadap.spectra.rowstackedspectra.RowStackedSpectra.binned_on_sky_area`
        and :func:`mean_sky_coordinates`.

        The arguments are passed directly to
        :func:`mean_sky_coordinates`, and the results are then passed
        to the base class function. The area of the fibers
        "beams" are all renormalized to :math:`\pi {\rm arcsec}^2` by
        the DRP.

        Args:
            bin_indx (array-like):
                An array with size :math:`N_{\rm spec}` that gives
                which spaxels were included in each bin. Valid bins
                have indices of :math:`\geq 0`.
            offset (:obj:`bool`, optional):
                Offset the coordinates in :attr:`xpos` and
                :attr:`ypos` by the difference between the IFU and
                OBJ coordinates in the header (see
                :func:`pointing_offset`). The coordinates are
                nominally defined with respect to the IFU pointing
                center, such that this offsets the coordinates to be
                centered on the galaxy.
            **kwargs:
                Passed directly to :func:`mean_sky_coordinates`.

        Returns:
            :obj:`tuple`: Two `numpy.ndarray`_ objects are returned.
            The first has the unique (non-negative) bin indices, and
            the second provides the on-sky area of that bin.
        """
        # TODO: May need a better way to do this. I.e., coordinates
        # will be need to figure out which spectra to bin, and this
        # recalculation of the fiducial coordinates runs the risk that
        # the binned area is not calculated using the same coordinates
        # used to determine which spectra went in each bin.
        x, y = self.mean_sky_coordinates(offset=offset, **kwargs)
        return super(MaNGARSS, self).binned_on_sky_area(bin_indx, x=x, y=y)

    @staticmethod
    def _parse_rectification_parameters(pixelscale, rlim, sigma, recenter, width_buffer):
        """
        Parse the datacube rectification parameters.

        Values that are None are set to their defaults as set by
        :class:`~mangadap.config.manga.MaNGAConfig`.

        Returns:
            :obj:`tuple`: The default or input values of arguments,
            in the same order.
        """
        pixelscale = MaNGAConfig.cube_pixelscale() if pixelscale is None else pixelscale
        rlim = MaNGAConfig.regrid_rlim() if rlim is None else rlim
        sigma = MaNGAConfig.regrid_sigma() if sigma is None else sigma
        recenter = MaNGAConfig.cube_recenter() if recenter is None else recenter
        width_buffer = MaNGAConfig.cube_width_buffer() if width_buffer is None else width_buffer
        return pixelscale, rlim, sigma, recenter, width_buffer

    def rectification_transfer_matrix(self, channel, pixelscale=None, rlim=None, sigma=None,
                                      recenter=None, width_buffer=None, quiet=False,
                                      rej_flag='3DREJECT'):
        """
        Construct the rectification transfer matrix.

        A simple wrapper for
        :func:`mangadap.spectra.rowstackedspectra.RowStackedSpectra.rectification_transfer_matrix`
        that defaults to the nominal rectification parameters for
        MaNGA datacubes. See this base class function for the
        argument descriptions.
        """
        _pixelscale, _rlim, _sigma, _recenter, _width_buffer \
                = MaNGARSS._parse_rectification_parameters(pixelscale, rlim, sigma, recenter,
                                                           width_buffer)
        return super(MaNGARSS,
            self).rectification_transfer_matrix(channel, _pixelscale, _rlim, _sigma,
                                                recenter=_recenter, width_buffer=_width_buffer,
                                                quiet=quiet, rej_flag='3DREJECT')

    def rectify_wavelength_plane(self, channel, pixelscale=None, rlim=None, sigma=None,
                                 recenter=None, width_buffer=None, quiet=False,
                                 rej_flag='3DREJECT', return_ivar=False, return_covar=False):
        """
        Rectify a wavelength channel into a monochromatic flux map.

        A simple wrapper for
        :func:`mangadap.spectra.rowstackedspectra.RowStackedSpectra.rectify_wavelength_plane`
        that defaults to the nominal rectification parameters for
        MaNGA datacubes. See this base class function for the
        argument descriptions.
        """
        _pixelscale, _rlim, _sigma, _recenter, _width_buffer \
                = MaNGARSS._parse_rectification_parameters(pixelscale, rlim, sigma, recenter,
                                                           width_buffer)
        return super(MaNGARSS, self).rectify_wavelength_plane(channel, pixelscale=_pixelscale,
                                                              rlim=_rlim, sigma=_sigma,
                                                              recenter=_recenter,
                                                              width_buffer=_width_buffer,
                                                              quiet=quiet, rej_flag='3DREJECT',
                                                              return_ivar=return_ivar,
                                                              return_covar=return_covar)

    def covariance_matrix(self, channel, pixelscale=None, rlim=None, sigma=None, recenter=None,
                          width_buffer=None, csr=False, quiet=False, rej_flag='3DREJECT'):
        """
        Construct the covariance matrix for a given wavelength channel.

        A simple wrapper for
        :func:`mangadap.spectra.rowstackedspectra.RowStackedSpectra.covariance_matrix`
        that defaults to the nominal rectification parameters for
        MaNGA datacubes. See this base class function for the
        argument descriptions.
        """
        _pixelscale, _rlim, _sigma, _recenter, _width_buffer \
                = MaNGARSS._parse_rectification_parameters(pixelscale, rlim, sigma, recenter,
                                                           width_buffer)
        return super(MaNGARSS, self).covariance_matrix(channel, pixelscale=_pixelscale, rlim=_rlim,
                                                       sigma=_sigma, recenter=_recenter,
                                                       width_buffer=_width_buffer, csr=csr,
                                                       quiet=quiet, rej_flag=rej_flag)

    def covariance_cube(self, channels=None, pixelscale=None, rlim=None, sigma=None, recenter=None,
                        width_buffer=None, csr=False, quiet=False, rej_flag='3DREJECT'):
        """
        Construct the covariance matrices for many wavelength channels.

        A simple wrapper for
        :func:`mangadap.spectra.rowstackedspectra.RowStackedSpectra.covariance_cube`
        that defaults to the nominal rectification parameters for
        MaNGA datacubes. See this base class function for the
        argument descriptions.
        """
        _pixelscale, _rlim, _sigma, _recenter, _width_buffer \
                = MaNGARSS._parse_rectification_parameters(pixelscale, rlim, sigma, recenter,
                                                           width_buffer)
        return super(MaNGARSS, self).covariance_cube(channels=channels, pixelscale=_pixelscale,
                                                     rlim=_rlim, sigma=_sigma, recenter=_recenter,
                                                     width_buffer=_width_buffer, csr=csr,
                                                     quiet=quiet, rej_flag=rej_flag)

    def instrumental_dispersion_plane(self, channel, dispersion_factor=None, pixelscale=None,
                                      rlim=None, sigma=None, recenter=None, width_buffer=None,
                                      quiet=False, rej_flag='3DREJECT'):
        """
        Construct the instrumental dispersion for a given wavelength channel.

        A simple wrapper for
        :func:`mangadap.spectra.rowstackedspectra.RowStackedSpectra.instrumental_dispersion_plane`
        that defaults to the nominal rectification parameters for
        MaNGA datacubes. See this base class function for the
        argument descriptions.
        """
        _pixelscale, _rlim, _sigma, _recenter, _width_buffer \
                = MaNGARSS._parse_rectification_parameters(pixelscale, rlim, sigma, recenter,
                                                           width_buffer)
        return super(MaNGARSS,
                    self).instrumental_dispersion_plane(channel,
                                                        dispersion_factor=dispersion_factor,
                                                        pixelscale=_pixelscale, rlim=_rlim,
                                                        sigma=_sigma, recenter=_recenter,
                                                        width_buffer=_width_buffer, quiet=quiet,
                                                        rej_flag=rej_flag)




# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
Provides the main wrapper function for the MaNGA DAP.

----

.. include license and copyright
.. include:: ../include/copy.rst

----

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""
import logging
import resource
import time
import os
import warnings

from IPython import embed

import numpy

import astropy.constants

from mangadap import __version__

from ..util.log import init_DAP_logging, module_logging, log_output
from ..datacube import DataCube
from ..config.analysisplan import AnalysisPlan
from ..proc.reductionassessments import ReductionAssessment
from ..proc.spatiallybinnedspectra import SpatiallyBinnedSpectra
from ..proc.stellarcontinuummodel import StellarContinuumModel
from ..proc.emissionlinemoments import EmissionLineMoments
from ..proc.emissionlinemodel import EmissionLineModel
from ..proc.spectralindices import SpectralIndices
from ..dapfits import construct_maps_file, construct_cube_file


def manga_dap(cube, plan, dbg=False, log=None, verbose=0):
    r"""
    Main wrapper function for the MaNGA DAP.

    This function is designed to be called once per datacube. The
    :class:`mangadap.config.analysisplan.AnalysisPlan` instance sets
    the types of analyses to perform on this observation. Each
    analysis plan results in a MAPS and model LOGCUBE file.

    The procedure is as follows.  For each plan in ``plan``:

        - Determine basic assessments of the data, including the S/N
          and spaxel coordinates. See
          :class:`mangadap.proc.reductionassessments.ReductionAssessment`.
        - Bin the spectra. See
          :class:`mangadap.proc.spatiallybinnedspectra.SpatiallyBinnedSpectra`.
        - Fit the stellar continuum for stellar kinematics. See
          :class:`mangadap.proc.stellarcontinuummodel.StellarContinuumModel`.
        - Measure the emission-line moments. See
          :class:`mangadap.proc.emissionlinemoments.EmissionLineMoments`.
        - Fit parameterized line profiles to the emission lines. See
          :class:`mangadap.proc.emissionlinemodel.EmissionLineModel`.
        - Subtract the fitted emission-line models and measure the
          spectral indices. See
          :class:`mangadap.proc.spectralindices.SpectralIndices`.
        - Construct the primary output files based on the plan
          results. See :func:`mangadap.dapfits.construct_maps_file`
          and :func:`mangadap.dapfits.construct_cube_file`.

    Verbose levels (still under development):
        0. Nothing but errors.
        1. Basic status updates. E.g., start and end of each block,
           minor progress within blocks.
        2. Block-level and sub-function updates.
        3. Same as above, with figures.

    Args:
        cube (:class:`mangadap.datacube.datacube.DataCube`):
            Datacube to analyze.
        plan (:class:`mangadap.config.analysisplan.AnalysisPlan`):
            Object that sets the analysis plans to implement and the output
            paths for DAP files.
        dbg (:obj:`bool`, optional):
            Flag to run the DAP in debug mode, see above; default is
            False. Limited use; still under development.
        log (:obj:`str`, optional):
            File name for log output, see above; no log file is
            produced by default.
        verbose (:obj:`int`, optional):
            Verbosity level, see above; default is 0.

    Returns:
        :obj:`int`: Status flag (under development; currently always
        0).
    """
    # Check input
    #  - Check that input cube type
    if not isinstance(cube, DataCube):
        raise TypeError('Input objects must be a subclass of DataCube.')
    # Check that the datacube can be analyzed by the DAP
    if not cube.can_analyze:
        raise ValueError('Cube metadata is insufficient for analysis.  See DataCube.can_analyze.')
    # Check that input plan type
    if not isinstance(plan, AnalysisPlan):
        raise TypeError('plan must be of type AnalysisPlan')

    # Initialize the logging objects
    init_DAP_logging(log)#, simple_warnings=False)

    # Start log
    loggers = module_logging(__name__, verbose)

    # Check the root output path exists, and create it if not.
    if not plan.analysis_path.exists():
        plan.analysis_path.mkdir(parents=True)

    log_output(loggers, 1, logging.INFO, '-'*50)
    log_output(loggers, 1, logging.INFO, '{0:^50}'.format('MaNGA Data Analysis Pipeline'))
    log_output(loggers, 1, logging.INFO, '-'*50)
    log_output(loggers, 1, logging.INFO, '     VERSION: {0}'.format(__version__))
    log_output(loggers, 1, logging.INFO, '       START: {0}'.format(
                                    time.strftime('%a %d %b %Y %H:%M:%S',time.localtime())))
    t = time.perf_counter()
    log_output(loggers, 1, logging.INFO, '  INPUT ROOT: {0}'.format(cube.directory_path))
    log_output(loggers, 1, logging.INFO, '   CUBE FILE: {0}'.format(cube.file_name))
    log_output(loggers, 1, logging.INFO, '     N PLANS: {0}'.format(plan.nplans))
    log_output(loggers, 1, logging.INFO, ' OUTPUT ROOT: {0}'.format(plan.analysis_path))
    log_output(loggers, 1, logging.INFO, '-'*50)

    status = 0

    # Iterate over plans:
    for i, key in enumerate(plan.keys()):
        # Construct some directories to ensure all the reference files
        # are placed in the same path.  The references files placed in
        # the common directory are:
        #   - ReductionAssessments
        #   - SpatiallyBinnedSpectra
        #   - StellarContinuumModel
        # These reference files are symlinked to each DAPTYPE reference
        # directory.  The reference files specific to each DAPTYPE are
        #   - EmissionLineMoments (although this could also go in the
        #   common directory)
        #   - EmissionLineModel
        #   - SpectralIndices
        common_dir = plan.common_path()
        method_dir = plan.method_path(plan_index=i)
        method_ref_dir = plan.method_path(plan_index=i, ref=True)

        log_output(loggers, 1, logging.INFO, '-'*50)
        log_output(loggers, 1, logging.INFO, f'{f"Plan {i+1}: {key}":^50}')
        log_output(loggers, 1, logging.INFO, '-'*50)
        log_output(loggers, 1, logging.INFO, f' METHOD KEY: {plan[key]["key"]}')
        log_output(loggers, 1, logging.INFO, f'COMMON PATH: {common_dir}')
        log_output(loggers, 1, logging.INFO, f'METHOD PATH: {method_dir}')

        #---------------------------------------------------------------
        # S/N Assessments: placed in the common/ directory
        #---------------------------------------------------------------
        rdxqa = None if plan.rdxqa[key] is None else \
                    ReductionAssessment(plan.rdxqa[key], cube, pa=cube.meta['pa'],
                                        ell=cube.meta['ell'], reff=cube.meta['reff'],
                                        output_path=common_dir, symlink_dir=method_ref_dir,
                                        loggers=loggers)

        #---------------------------------------------------------------
        # Spatial Binning: placed in the common/ directory
        #---------------------------------------------------------------
        binned_spectra = None if plan.binning[key] is None else \
                    SpatiallyBinnedSpectra(plan.binning[key], cube, rdxqa, reff=cube.meta['reff'],
                                           ebv=cube.meta['ebv'], output_path=common_dir,
                                           symlink_dir=method_ref_dir, loggers=loggers)

        #---------------------------------------------------------------
        # Stellar Continuum Fit: placed in the common/ directory
        #---------------------------------------------------------------
        # TODO: Allow tpl_hardcopy to be an input parameter?
        stellar_continuum = None if plan.continuum[key] is None else \
                    StellarContinuumModel(plan.continuum[key], binned_spectra,
                                          guess_vel=cube.meta['vel'], guess_sig=cube.meta['vdisp'],
                                          output_path=common_dir, symlink_dir=method_ref_dir,
                                          loggers=loggers)

        #---------------------------------------------------------------
        # For both the emission-line measurements (moments and model):
        # To use the NSA redshift for all moments:
        #   - set redshift=nsa_redshift
        # To use the stellar kinematics to set the redshift:
        #   - set redshift=None
        #---------------------------------------------------------------
        # Emission-line Moment measurements: placed in the DAPTYPE/ref/
        # directory
        #---------------------------------------------------------------
#        warnings.filterwarnings('error', message='Warning: converting a masked element to nan.')
        # TODO: enable artifact_path and bandpass_path?
        emission_line_moments = None if plan.elmom[key] is None else \
                    EmissionLineMoments(plan.elmom[key], binned_spectra,
                                        stellar_continuum=stellar_continuum,
                                        redshift=cube.meta['z'], output_path=method_ref_dir,
                                        loggers=loggers)

        #---------------------------------------------------------------
        # Emission-line Fit: placed in the DAPTYPE/ref/ directory
        #---------------------------------------------------------------
        # To use a fixed velocity dispersion for the initial guess, e.g.:
        #   - set dispersion=100.0
        # To use the stellar kinematics as the initial guess:
        #   - set dispersion=None
        # The provided redshift is only the initial guess for the
        # emission-line model (it's FIXED for the moment measurements
        # above).
        emission_line_model = None if plan.elfit[key] is None else \
                   EmissionLineModel(plan.elfit[key], binned_spectra,
                                     stellar_continuum=stellar_continuum,
                                     emission_line_moments=emission_line_moments, dispersion=100.0,
                                     minimum_error=numpy.finfo(numpy.float32).eps,
                                     output_path=method_ref_dir, loggers=loggers)

        #---------------------------------------------------------------
        # If requested by the emission-line moments method, remeasure
        # the moments after the emission-line modeling.  This will
        # produce a new reference file that will have a different name
        # than the one produced above.  This is placed in the
        # DAPTYPE/ref/ directory.
        #---------------------------------------------------------------
        if emission_line_moments is not None \
                and emission_line_moments.database['redo_postmodeling']:
            emission_line_moments.measure(binned_spectra, stellar_continuum=stellar_continuum,
                                          emission_line_model=emission_line_model,
                                          output_path=method_ref_dir, loggers=loggers)

        #---------------------------------------------------------------
        # Spectral-Index Measurements: placed in the DAPTYPE/ref/
        # directory
        #---------------------------------------------------------------
        spectral_indices = None if plan.sindx[key] is None else \
                    SpectralIndices(plan.sindx[key], binned_spectra, redshift=cube.meta['z'],
                                    stellar_continuum=stellar_continuum,
                                    emission_line_model=emission_line_model,
                                    output_path=method_ref_dir, loggers=loggers)

        #-------------------------------------------------------------------
        # Construct the main output file(s)
        #-------------------------------------------------------------------
        construct_maps_file(cube, method=plan[key]['key'], rdxqa=rdxqa,
                            binned_spectra=binned_spectra, stellar_continuum=stellar_continuum,
                            emission_line_moments=emission_line_moments,
                            emission_line_model=emission_line_model,
                            spectral_indices=spectral_indices, redshift=cube.meta['z'],
                            output_path=method_dir, overwrite=True, loggers=loggers,
                            single_precision=True)

        construct_cube_file(cube, method=plan[key]['key'], rdxqa=rdxqa,
                            binned_spectra=binned_spectra, stellar_continuum=stellar_continuum,
                            emission_line_moments=emission_line_moments,
                            emission_line_model=emission_line_model,
                            spectral_indices=spectral_indices, output_path=method_dir,
                            overwrite=True, loggers=loggers, single_precision=True)

        # Mark for next garbage collection.
        # TODO: Not sure this is useful because these variables may
        # just get immediately defined again in the next iteration.
        del spectral_indices
        del emission_line_model
        del emission_line_moments
        del stellar_continuum
        del binned_spectra
        del rdxqa

    #-------------------------------------------------------------------
    # End log
    #-------------------------------------------------------------------
    log_output(loggers, 1, logging.INFO, '-'*50)
    log_output(loggers, 1, logging.INFO, 'EXECUTION SUMMARY')
    log_output(loggers, 1, logging.INFO, '-'*50)
    log_output(loggers, 1, logging.INFO,
               f'    FINISH: {time.strftime("%a %d %b %Y %H:%M:%S",time.localtime())}')
    if time.perf_counter() - t < 60:
        tstr = f'  DURATION: {time.perf_counter() - t:.5f} sec'
    elif time.perf_counter() - t < 3600:
        tstr = f'  DURATION: {(time.perf_counter() - t)/60.:.5f} min'
    else:
        tstr = f'  DURATION: {(time.perf_counter() - t)/3600.:.5f} hr'
    log_output(loggers, 1, logging.INFO, tstr)

    mem = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss \
            + resource.getrusage(resource.RUSAGE_CHILDREN).ru_maxrss
    if mem < 1024:
        mstr = f'MAX MEMORY: {mem:.4f} bytes'
    elif mem < 1024*1024:
        mstr = f'MAX MEMORY: {mem/1024:.4f} Kbytes'
    elif mem < 1024*1024*1024:
        mstr = f'MAX MEMORY: {mem/1024/1024:.4f} Mbytes'
    else:
        mstr = f'MAX MEMORY: {mem/1024/1024/1024:.4f} Gbytes'
    log_output(loggers, 1, logging.INFO, mstr)
    log_output(loggers, 1, logging.INFO, '-'*50)
    log_output(loggers, 1, logging.INFO, '-'*50)

    return status





# -*- coding: utf-8 -*-
"""
Utility functions for parameter sets.

----

.. include license and copyright
.. include:: ../include/copy.rst

----

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""
import numpy 


def _eval_ignore():
    """Provides a list of strings that should not be evaluated."""
    return [ 'open', 'file', 'dict', 'list', 'tuple' ]


def recursive_dict_evaluate(d):
    """
    Recursively run `eval`_ on each element of the provided
    dictionary.

    A raw read of a configuration file with `configobj.ConfigObj`_
    results in a dictionary that contains strings or lists of strings.
    However, when assigning the values for the various
    :class:`mangadap.par.parset.ParSet` objects, the
    :func:`mangadap.par.parset.ParSet.from_dict` methods expect the
    dictionary values to have the appropriate type.  E.g., the
    `configobj.ConfigObj`_ will have something like ``d['foo'] = '1'``,
    when the :func:`mangadap.par.parset.ParSet.from_dict` method expects
    the value to be an integer (``d['foo'] = 1``).

    This function tries to evaluate *all* dictionary values, except for
    those listed by the :func:`_eval_ignore` function.  Any value in
    this list or where::

        eval(d[k]) for k in d.keys()

    raises an exception and is returned as the original string.

    Args:
        d (:obj:`dict`):
            Dictionary of values to evaluate

    Returns:
        :obj:`dict`: Identical to input dictionary, but with all string
        values replaced with the result of ``eval(d[k])`` for all ``k``
        in ``d.keys()``.
    """
    #ignore = _eval_ignore()
    for k in d.keys():
        if isinstance(d[k], dict):
           d[k] = recursive_dict_evaluate(d[k])
        elif isinstance(d[k], list):
            replacement = []
            for v in d[k]:
                if v in _eval_ignore():
                    replacement += [ v ]
                else:
                    try:
                        replacement += [ eval(v) ]
                    except:
                        replacement += [ v ]
            d[k] = replacement
        else:
            try:
                d[k] = eval(d[k]) if d[k] not in _eval_ignore() else d[k]
            except:
                pass

    return d


def recursive_dict_str_to_None(d):
    """
    Recursively run any string values of 'None' in a dictionary to be explicitly
    None.

    Args:
        d (:obj:`dict`):
            Dictionary of values to evaluate

    Returns:
        :obj:`dict`: Identical to input dictionary, but with all string values
        of 'None' set to None explicitly.  **The input dictionary is altered
        directly.**
    """
    for k in d.keys():
        if isinstance(d[k], dict):
           d[k] = recursive_dict_evaluate(d[k])
        elif isinstance(d[k], list):
            replacement = []
            for v in d[k]:
                replacement += [v if v != 'None' else None]
            d[k] = replacement
        elif d[k] == 'None':
            d[k] = None
    return d


def get_parset_list(cfg, pk, parsetclass):
    """
    Create a list of :class:`mangadap.par.parset.ParSet` objects based
    on a root keyword for a set of defined groups in the configuration
    file.
    
    For example, the a parameter group could allow for a list of parset
    objects instead of a single one.  This function parses the provided
    configuration object (``cfg``) to find any sections with a string
    keyword (`pk`) as its root.  The remainder of the section name must
    be able to be converted to an integer and the section itself must be
    able to setup an instance of `parsetclass`.  The sections must be
    number sequentially from 1..N.  

    Args:
        cfg (`configobj.ConfigObj`_, :obj:`dict`):
            The top-level configuration that defines a list of
            sub-ParSets.
        pk (:obj:`str`):
            The root of the keywords used to set a list of sub-ParSets.
        parsetclass (:class:`mangadap.par.parset.ParSet`):
            The class used to construct each element in the list of
            parameter subsets.  The class **must** have a ``from_dict``
            method that instantiates the
            :class:`mangadap.par.parset.ParSet` based on the provided
            subsection/subdict from ``cfg``.

    Returns:
        :obj:`list`: A list of instances of ``parsetclass`` parsed from
        the provided configuration data.

    Raises:
        ValueError:
            Raised if the indices of the subsections are not sequential
            and 1-indexed.
    """
    # Get the full list of keys
    k = cfg.keys()

    # Iterate through the list of keys to find the appropriate sub
    # parameter sets and their order.
    par = []
    order = []
    for _k in k:
        if _k == pk and cfg[_k] is None:
            continue
        if pk in _k:
            try:
                # Get the order for this subgroup (e.g., 2 for
                # 'detector2'
                order += [ int(_k.replace(pk,'')) ]
                # And instantiate the parameter set
                par += [ parsetclass.from_dict(cfg[_k]) ]
            except:
                continue

    if len(par) > 0:
        # Make sure the instances are correctly sorted and sequential
        srt = numpy.argsort(order)
        if numpy.any(numpy.array(order)[srt]-1 != numpy.arange(order[srt[-1]])):
            raise ValueError('Parameter set series must be sequential and 1-indexed.')
        # Return the sorted instances
        return [par[i] for i in srt]

    # No such subsets were defined, so return a null result
    return None


